# this defines the public API of mpi4jax
# all other intialization is taking place in _src/__init__.py

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

from ._src import (  # noqa: E402
    allgather,
    allreduce,
    alltoall,
    barrier,
    bcast,
    gather,
    recv,
    reduce,
    scan,
    scatter,
    send,
    sendrecv,
    has_cuda_support,
)

__all__ = [
    "allgather",
    "allreduce",
    "alltoall",
    "barrier",
    "bcast",
    "gather",
    "recv",
    "reduce",
    "scan",
    "scatter",
    "send",
    "sendrecv",
    "has_cuda_support",
]



import pytest

from mpi4py import MPI

import jax
import jax.numpy as jnp

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv():
    from mpi4jax import sendrecv

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    other = 1 - rank

    res, token = sendrecv(arr, arr, source=other, dest=other)

    assert jnp.array_equal(res, jnp.ones_like(arr) * other)
    assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_status():
    from mpi4jax import sendrecv

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    other = 1 - rank

    status = MPI.Status()
    res, token = sendrecv(arr, arr, source=other, dest=other, status=status)

    assert jnp.array_equal(res, jnp.ones_like(arr) * other)
    assert jnp.array_equal(_arr, arr)
    assert status.Get_source() == other


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_status_jit():
    from mpi4jax import sendrecv

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    other = 1 - rank

    status = MPI.Status()
    res = jax.jit(
        lambda x, y: sendrecv(x, y, source=other, dest=other, status=status)[0]
    )(arr, arr)

    assert jnp.array_equal(res, jnp.ones_like(arr) * other)
    assert jnp.array_equal(_arr, arr)
    assert status.Get_source() == other


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_scalar():
    from mpi4jax import sendrecv

    arr = 1 * rank
    _arr = arr

    other = 1 - rank

    res, token = sendrecv(arr, arr, source=other, dest=other)

    assert jnp.array_equal(res, jnp.ones_like(arr) * other)
    assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_jit():
    from mpi4jax import sendrecv

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    other = 1 - rank

    res = jax.jit(lambda x, y: sendrecv(x, y, source=other, dest=other)[0])(arr, arr)

    assert jnp.array_equal(res, jnp.ones_like(arr) * other)
    assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_scalar_jit():
    from mpi4jax import sendrecv

    arr = 1 * rank
    _arr = arr

    other = 1 - rank

    res = jax.jit(lambda x, y: sendrecv(x, y, source=other, dest=other)[0])(arr, arr)

    assert jnp.array_equal(res, jnp.ones_like(arr) * other)
    assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_vmap():
    from mpi4jax import sendrecv

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    other = 1 - rank

    res = sendrecv(arr, arr, source=other, dest=other)[0]

    def fun(x, y):
        return sendrecv(x, y, source=other, dest=other)[0]

    vfun = jax.vmap(fun, in_axes=(0, 0))
    res = vfun(_arr, arr)

    assert jnp.array_equal(res, jnp.ones_like(arr) * other)
    assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_grad():
    from mpi4jax import sendrecv

    arr = jnp.ones((3, 2)) * (rank + 1)
    _arr = arr.copy()

    other = 1 - rank

    def f(x):
        x, token = sendrecv(x, x, source=other, dest=other)
        x = x * (rank + 1)
        return x.sum()

    res = jax.grad(f)(arr)

    assert jnp.array_equal(res, jnp.ones_like(arr) * (other + 1))
    assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_grad_2():
    from mpi4jax import sendrecv

    arr = jnp.ones((3, 2)) * (rank + 1)
    _arr = arr.copy()

    other = 1 - rank

    def f(x):
        x, token = sendrecv(x, x, source=other, dest=other)
        x = x * (rank + 1) * 5
        x, token = sendrecv(x, x, source=other, dest=other, token=token)
        x = x * (rank + 1) ** 2
        return x.sum()

    res = jax.grad(f)(arr)

    solution = (rank + 1) ** 2 * (other + 1) * 5
    print("solution is ", solution)
    assert jnp.array_equal(res, jnp.ones_like(arr) * solution)
    assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_jacfwd():
    from mpi4jax import sendrecv

    arr = jnp.ones((3, 2)) * (rank + 1)
    # _arr = arr.copy()

    other = 1 - rank

    def f(x):
        x, token = sendrecv(x, x, source=other, dest=other)
        x = x * (rank + 1)
        return x.sum()

    with pytest.raises(RuntimeError):
        jax.jacfwd(f)(arr)

    # assert jnp.array_equal(res, jnp.ones_like(arr) * (other + 1))
    # assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_jacrev():
    from mpi4jax import sendrecv

    arr = jnp.ones((3, 2)) * (rank + 1)
    _arr = arr.copy()

    other = 1 - rank

    def f(x):
        x, token = sendrecv(x, x, source=other, dest=other)
        x = x * (rank + 1)
        return x.sum()

    res = jax.jacrev(f)(arr)

    assert jnp.array_equal(res, jnp.ones_like(arr) * (other + 1))
    assert jnp.array_equal(_arr, arr)



import pytest


def test_envvar_parsing():
    from mpi4jax._src.decorators import _is_falsy, _is_truthy

    assert _is_truthy("1")
    assert not _is_falsy("1")

    assert not _is_truthy("false")
    assert _is_falsy("false")

    assert not _is_truthy("foo")
    assert not _is_falsy("foo")


def test_ensure_gpu_ext(monkeypatch):
    from mpi4jax._src import xla_bridge
    from mpi4jax._src.decorators import ensure_gpu_ext

    with monkeypatch.context() as m:
        m.setattr(xla_bridge, "HAS_GPU_EXT", False)

        with pytest.raises(ImportError) as excinfo:
            ensure_gpu_ext()

        assert "GPU extensions could not be imported" in str(excinfo.value)



from .._src.collective_ops.allgather import mpi_allgather_p
from .._src.collective_ops.allreduce import mpi_allreduce_p
from .._src.collective_ops.alltoall import mpi_alltoall_p
from .._src.collective_ops.barrier import mpi_barrier_p
from .._src.collective_ops.bcast import mpi_bcast_p
from .._src.collective_ops.gather import mpi_gather_p
from .._src.collective_ops.recv import mpi_recv_p
from .._src.collective_ops.reduce import mpi_reduce_p
from .._src.collective_ops.scan import mpi_scan_p
from .._src.collective_ops.send import mpi_send_p
from .._src.collective_ops.scatter import mpi_scatter_p
from .._src.collective_ops.sendrecv import mpi_sendrecv_p


token_override_registry = {}


def mpi_allgather_token_override(in_args, new_token, comm):
    x, _ = in_args
    return mpi_allgather_p.bind(x, new_token, comm=comm)


token_override_registry[mpi_allgather_p] = mpi_allgather_token_override


def mpi_allreduce_token_override(in_args, new_token, op, comm, transpose):
    x, _ = in_args
    return mpi_allreduce_p.bind(x, new_token, op=op, comm=comm, transpose=transpose)


token_override_registry[mpi_allreduce_p] = mpi_allreduce_token_override


def mpi_alltoall_token_override(in_args, new_token, comm):
    x, _ = in_args
    return mpi_alltoall_p.bind(x, new_token, comm=comm)


token_override_registry[mpi_alltoall_p] = mpi_alltoall_token_override


def mpi_barrier_token_override(in_args, new_token, comm):
    return (mpi_barrier_p.bind(new_token, comm=comm),)


token_override_registry[mpi_barrier_p] = mpi_barrier_token_override


def mpi_bcast_token_override(in_args, new_token, root, comm):
    x, _ = in_args
    return mpi_bcast_p.bind(x, new_token, root=root, comm=comm)


token_override_registry[mpi_bcast_p] = mpi_bcast_token_override


def mpi_gather_token_override(in_args, new_token, root, comm):
    x, _ = in_args
    return mpi_gather_p.bind(x, new_token, root=root, comm=comm)


token_override_registry[mpi_gather_p] = mpi_gather_token_override


def mpi_recv_token_override(in_args, new_token, source, tag, comm, status):
    x, _ = in_args
    return mpi_recv_p.bind(
        x, new_token, source=source, tag=tag, comm=comm, status=status
    )


token_override_registry[mpi_recv_p] = mpi_recv_token_override


def mpi_reduce_token_override(in_args, new_token, op, root, comm):
    x, _ = in_args
    return mpi_reduce_p.bind(x, new_token, op=op, root=root, comm=comm)


token_override_registry[mpi_reduce_p] = mpi_reduce_token_override


def mpi_scan_token_override(in_args, new_token, op, comm):
    x, _ = in_args
    return mpi_scan_p.bind(x, new_token, op=op, comm=comm)


token_override_registry[mpi_scan_p] = mpi_scan_token_override


def mpi_scatter_token_override(in_args, new_token, root, comm):
    x, _ = in_args
    return mpi_scatter_p.bind(x, new_token, root=root, comm=comm)


token_override_registry[mpi_scatter_p] = mpi_scatter_token_override


def mpi_send_token_override(in_args, new_token, dest, tag, comm):
    x, _ = in_args
    return (mpi_send_p.bind(x, new_token, dest=dest, tag=tag, comm=comm),)


token_override_registry[mpi_send_p] = mpi_send_token_override


def mpi_sendrecv_token_override(
    in_args, new_token, source, dest, sendtag, recvtag, comm, status, _must_transpose
):
    sendbuff, recvbuff, _ = in_args
    return mpi_sendrecv_p.bind(
        sendbuff,
        recvbuff,
        new_token,
        source=source,
        dest=dest,
        sendtag=sendtag,
        recvtag=recvtag,
        comm=comm,
        status=status,
        _must_transpose=_must_transpose,
    )


token_override_registry[mpi_sendrecv_p] = mpi_sendrecv_token_override



import jax
import jax.numpy as jnp

from mpi4py import MPI
from mpi4jax import allreduce

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()


seed = 123
m, n = 5, 60

k1, k2, k3, k4, k5 = jax.random.split(jax.random.PRNGKey(seed), 5)
A = jax.random.uniform(k1, (m, n))
x = jax.random.uniform(k2, (n,))
y = jax.random.uniform(k3, (m,))
v = jax.random.uniform(k4, (n,))
vprime = jax.random.uniform(k5, (m,))


assert n % size == 0
n_local = n // size
start_local = n_local * rank
end_local = start_local + n_local

A_local = A[:, start_local:end_local]
x_local = x[start_local:end_local]
v_local = v[start_local:end_local]


# expected results:
Ax = A @ x
ATy_local = (A.T @ y)[start_local:end_local]

Av = A @ v
ATvprime_local = (A.T @ vprime)[start_local:end_local]


def allreduce_sum(x):
    res, _ = allreduce(x, op=MPI.SUM, comm=MPI.COMM_WORLD)
    return res


# x is distributed across ranks
# cols of A are distributed across ranks
# result is same on all ranks
def matvec_mpi(A_local, x_local):
    return allreduce_sum(A_local @ x_local)


# y is same on all ranks
# rows of A.T are distributed across ranks
# result is distributed across ranks
def matvec_transpose_mpi(A_local, y_global):
    return A_local.T @ transpose(allreduce_sum, y_global)(y_global)


def mv(x_local):
    return matvec_mpi(A_local, x_local)


def mvT(y):
    return matvec_transpose_mpi(A_local, y)


def transpose(f, x):
    def fT(y):
        return jax.linear_transpose(f, x)(y)[0]

    return fT


# tests


def test_matvec():
    res = mv(x_local)
    assert jnp.allclose(Ax, res)


def test_matvec_jit():
    res = jax.jit(mv)(x_local)
    assert jnp.allclose(Ax, res)


def test_matvecT():
    res = mvT(y)
    assert jnp.allclose(ATy_local, res)


def test_matvecT_jit():
    res = jax.jit(mvT)(y)
    assert jnp.allclose(ATy_local, res)


def test_matvec_transpose():
    lt = transpose(mv, x_local)
    res = lt(y)
    assert jnp.allclose(ATy_local, res)


def test_matvec_transpose_jit():
    lt = transpose(mv, x_local)
    res = jax.jit(lt)(y)
    assert jnp.allclose(ATy_local, res)


def test_matvecT_transpose():
    ltT = transpose(mvT, y)
    res = ltT(x_local)
    assert jnp.allclose(Ax, res)


def test_matvecT_transpose_jit():
    ltT = transpose(mvT, y)
    res = jax.jit(ltT)(x_local)
    assert jnp.allclose(Ax, res)


def test_matvec_transpose2():
    lt = transpose(mv, x_local)
    ltlt = transpose(lt, y)
    res = ltlt(x_local)
    assert jnp.allclose(Ax, res)


def test_matvec_transpose2_jit():
    lt = transpose(mv, x_local)
    ltlt = transpose(lt, y)
    res = jax.jit(ltlt)(x_local)
    assert jnp.allclose(Ax, res)


def test_matvecT_transpose2():
    ltT = transpose(mvT, y)
    ltltT = transpose(ltT, x_local)
    res = ltltT(y)
    assert jnp.allclose(ATy_local, res)


def test_matvecT_transpose2_jit():
    ltT = transpose(mvT, y)
    ltltT = transpose(ltT, x_local)
    res = jax.jit(ltltT)(y)
    assert jnp.allclose(ATy_local, res)


def test_matvec_transpose3():
    lt = transpose(mv, x_local)
    ltlt = transpose(lt, y)
    ltltlt = transpose(ltlt, x_local)
    res = ltltlt(y)
    assert jnp.allclose(ATy_local, res)


def test_matvec_transpose3_jit():
    lt = transpose(mv, x_local)
    ltlt = transpose(lt, y)
    ltltlt = transpose(ltlt, x_local)
    res = jax.jit(ltltlt)(y)
    assert jnp.allclose(ATy_local, res)


def test_matvecT_transpose3():
    ltT = transpose(mvT, y)
    ltltT = transpose(ltT, x_local)
    ltltltT = transpose(ltltT, y)
    res = ltltltT(x_local)
    assert jnp.allclose(Ax, res)


def test_matvecT_transpose3_jit():
    ltT = transpose(mvT, y)
    ltltT = transpose(ltT, x_local)
    ltltltT = transpose(ltltT, y)
    res = jax.jit(ltltltT)(x_local)
    assert jnp.allclose(Ax, res)


def test_matvec_jvp():
    res, jvp = jax.jvp(mv, (x_local,), (v_local,))
    assert jnp.allclose(Ax, res)
    assert jnp.allclose(Av, jvp)


def test_matvec_jvp_jit():
    res, jvp = jax.jit(lambda x, v: jax.jvp(mv, (x,), (v,)))(x_local, v_local)
    assert jnp.allclose(Ax, res)
    assert jnp.allclose(Av, jvp)


def test_matvecT_jvp():
    res, jvp = jax.jvp(mvT, (y,), (vprime,))
    assert jnp.allclose(ATy_local, res)
    assert jnp.allclose(ATvprime_local, jvp)


def test_matvecT_jvp_jit():
    res, jvp = jax.jit(lambda y, v: jax.jvp(mvT, (y,), (v,)))(y, vprime)
    assert jnp.allclose(ATy_local, res)
    assert jnp.allclose(ATvprime_local, jvp)


def test_matvec_vjp():
    res, vjp_fun = jax.vjp(mv, x_local)
    (vjp,) = vjp_fun(vprime)
    assert jnp.allclose(Ax, res)
    assert jnp.allclose(ATvprime_local, vjp)


def test_matvec_vjp_jit():
    def f(x, v):
        res, vjp_fun = jax.vjp(mv, x)
        (vjp,) = vjp_fun(v)
        return res, vjp

    res, vjp = jax.jit(f)(x_local, vprime)
    assert jnp.allclose(Ax, res)
    assert jnp.allclose(ATvprime_local, vjp)


def test_matvecT_vjp():
    res, vjp_fun = jax.vjp(mvT, y)
    (jvp,) = vjp_fun(v_local)
    assert jnp.allclose(ATy_local, res)
    assert jnp.allclose(Av, jvp)


def test_matvecT_vjp_jit():
    def f(y, v):
        res, vjp_fun = jax.vjp(mvT, y)
        (jvp,) = vjp_fun(v)
        return res, jvp

    res, jvp = jax.jit(f)(y, v_local)
    assert jnp.allclose(ATy_local, res)
    assert jnp.allclose(Av, jvp)



def pytest_report_header(config):
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    return [
        f"MPI vendor: {MPI.get_vendor()}",
        f"MPI rank: {comm.Get_rank()}",
        f"MPI size: {comm.Get_size()}",
    ]


def pytest_configure(config):
    # don't hog memory if running on GPU
    import os

    os.environ["XLA_PYTHON_CLIENT_PREALLOCATE"] = "false"



import functools
import inspect

import numpy as _np
from jax.core import Tracer


def enforce_types(**types):
    """Decorator that enforces given argument types at runtime.

    Throws a TypeError if an invalid type is passed.

    If argument type is a subclass of np.generic (e.g. np.integer or np.floating),
    allow all sub-dtypes (np.integer allows int, np.uint64, np.int16, ...).

    Example:

        >>> @enforce_types(
        ...     foo=(np.integer, type(None)),
        ...     bar=str
        ... )
        ... def func(x, foo, bar):
        ...     print(bar)

        >>> func(1, 2, 'hi there!')
        hi there!

        >>> func(1, 2, 3)
        TypeError: func got unexpected type for argument "bar" (expected: str, got:
        <class 'int'>)

    """

    def decorator(function):
        func_name = function.__name__
        func_sig = inspect.signature(function)

        for t in types:
            # make sure given types are actually parameters of decorated function
            if t not in func_sig.parameters:
                raise ValueError(
                    f"enforce_types decorator for {func_name} "
                    f'got unexpected argument "{t}"'
                )

            # make sure types are iterable
            if not isinstance(types[t], (tuple, list)):
                types[t] = (types[t],)

        @functools.wraps(function)
        def wrapped(*args, **kwargs):
            # parse passed args
            bound_args = func_sig.bind(*args, **kwargs)
            bound_args.apply_defaults()

            # check arguments one by one
            for arg, val in bound_args.arguments.items():
                if arg not in types:
                    continue

                arg_types = types[arg]

                # check types one by one
                check_passed = False
                for t in arg_types:
                    if issubclass(t, _np.generic):
                        check_passed |= _np.issubdtype(type(val), t)
                    else:
                        check_passed |= isinstance(val, t)

                if not check_passed:
                    readable_arg_types = [t.__qualname__ for t in arg_types]
                    if len(readable_arg_types) == 1:
                        readable_arg_types = readable_arg_types[0]

                    extra_message = ""
                    if isinstance(val, Tracer):
                        extra_message = (
                            "\n\nAn abstract tracer was passed where a concrete value "
                            "is expected. Try using the static_argnums argument "
                            "of jax.jit."
                        )

                    raise TypeError(
                        f'{func_name} got unexpected type for argument "{arg}" '
                        f"(expected: {readable_arg_types}, got: {type(val)})."
                        f"{extra_message}"
                    )

            return function(*args, **kwargs)

        return wrapped

    return decorator



import numpy as _np
from mpi4py import MPI as _MPI

from jax import abstract_arrays, core
from jax.core import Primitive, Tracer, Token
from jax.lax import create_token

from jax.interpreters import mlir
import jaxlib.mlir.ir as ir

from ..utils import (
    HashableMPIType,
    default_primitive_impl,
    to_dtype_handle,
    to_mpi_handle,
    unpack_hashable,
    wrap_as_hashable,
    as_mhlo_constant,
    get_default_layouts,
    effect,
)
from ..jax_compat import hlo_custom_call, token_type
from ..decorators import translation_rule_cpu, translation_rule_gpu
from ..validation import enforce_types
from ..comm import get_default_comm


# The Jax primitive
mpi_reduce_p = Primitive("reduce_mpi")  # Create the primitive
mpi_reduce_impl = default_primitive_impl(mpi_reduce_p)


# This function applies the primitive to an AST
@enforce_types(
    op=(_MPI.Op, HashableMPIType),
    root=(_np.integer),
    comm=(type(None), _MPI.Intracomm, HashableMPIType),
    token=(type(None), Token, Tracer),
)
def reduce(x, op, root, *, comm=None, token=None):
    """Perform a reduce operation.

    Arguments:
        x: Array or scalar input to send.
        op (mpi4py.MPI.Op): The reduction operator (e.g :obj:`mpi4py.MPI.SUM`).
        root (int): Rank of the root MPI process.
        comm (mpi4py.MPI.Comm): The MPI communicator to use (defaults to
            a clone of :obj:`COMM_WORLD`).
        token (Token): XLA token to use to ensure correct execution order.
            If not given, a new token is generated.

    Returns:
        Tuple[DeviceArray, Token]:
            - Result of the reduce operation on root process, otherwise
              unmodified input.
            - A new, modified token, that depends on this operation.

    """
    if token is None:
        token = create_token(x)

    if comm is None:
        comm = get_default_comm()

    rank = comm.Get_rank()

    op = wrap_as_hashable(op)
    comm = wrap_as_hashable(comm)
    res, token = mpi_reduce_p.bind(x, token, op=op, root=root, comm=comm)

    if rank != root:
        return (x, token)

    return (res, token)


# This function compiles the operation
@translation_rule_cpu
def mpi_reduce_xla_encode_cpu(ctx, x, token, op, root, comm):
    op = unpack_hashable(op)
    comm = unpack_hashable(comm)

    x_aval, *_ = ctx.avals_in
    x_nptype = x_aval.dtype

    x_type = ir.RankedTensorType(x.type)
    dtype = x_type.element_type
    dims = x_type.shape

    # compute total number of elements in array
    nitems = _np.prod(dims, dtype=int)

    dtype_handle = to_dtype_handle(x_nptype)

    # output is only used on root, so prevent memory allocation
    rank = comm.Get_rank()

    if rank != root:
        dims = (0,)

    out_types = [
        ir.RankedTensorType.get(dims, dtype),
        *token_type(),
    ]

    operands = (
        as_mhlo_constant(nitems, _np.intc),
        x,
        as_mhlo_constant(to_mpi_handle(op), _np.uintp),
        as_mhlo_constant(root, _np.intc),
        as_mhlo_constant(to_mpi_handle(comm), _np.uintp),
        as_mhlo_constant(dtype_handle, _np.uintp),
        token,
    )

    return hlo_custom_call(
        b"mpi_reduce",
        out_types=out_types,
        operands=operands,
        operand_layouts=get_default_layouts(operands),
        result_layouts=get_default_layouts(out_types),
        has_side_effect=True,
    )


@translation_rule_gpu
def mpi_reduce_xla_encode_gpu(ctx, x, token, op, root, comm):
    from ..xla_bridge.mpi_xla_bridge_gpu import build_reduce_descriptor

    op = unpack_hashable(op)
    comm = unpack_hashable(comm)

    x_aval, *_ = ctx.avals_in
    x_nptype = x_aval.dtype

    x_type = ir.RankedTensorType(x.type)
    dtype = x_type.element_type
    dims = x_type.shape

    # compute total number of elements in array
    nitems = _np.prod(dims, dtype=int)

    dtype_handle = to_dtype_handle(x_nptype)

    # output is only used on root, so prevent memory allocation
    rank = comm.Get_rank()

    if rank != root:
        dims = (0,)

    out_types = [
        ir.RankedTensorType.get(dims, dtype),
        *token_type(),
    ]

    operands = (
        x,
        token,
    )

    descriptor = build_reduce_descriptor(
        nitems,
        to_mpi_handle(op),
        root,
        to_mpi_handle(comm),
        dtype_handle,
    )

    return hlo_custom_call(
        b"mpi_reduce",
        out_types=out_types,
        operands=operands,
        operand_layouts=get_default_layouts(operands),
        result_layouts=get_default_layouts(out_types),
        has_side_effect=True,
        backend_config=descriptor,
    )


# This function evaluates only the shapes during AST construction
def mpi_reduce_abstract_eval(xs, token, op, root, comm):
    comm = unpack_hashable(comm)
    rank = comm.Get_rank()

    if rank != root:
        dims = (0,)
    else:
        dims = xs.shape

    return (
        abstract_arrays.ShapedArray(dims, xs.dtype),
        core.abstract_token,
    ), {effect}


mpi_reduce_p.multiple_results = True
mpi_reduce_p.def_impl(mpi_reduce_impl)
mpi_reduce_p.def_effectful_abstract_eval(mpi_reduce_abstract_eval)

# assign to the primitive the correct encoder
mlir.register_lowering(mpi_reduce_p, mpi_reduce_xla_encode_cpu, platform="cpu")
mlir.register_lowering(mpi_reduce_p, mpi_reduce_xla_encode_gpu, platform="cuda")



from cpython.pycapsule cimport PyCapsule_New

from libc.stdint cimport uintptr_t
from libc.stdlib cimport malloc, free

from mpi4py.libmpi cimport (
    MPI_Comm,
    MPI_Comm_rank,
    MPI_Comm_size,
    MPI_Datatype,
    MPI_Op,
    MPI_Status,
    MPI_Type_size,
)

from .cuda_runtime_api cimport (
    cudaGetErrorName,
    cudaGetErrorString,
    cudaError_t,
    cudaMemcpy,
    cudaMemcpyDeviceToDevice,
    cudaMemcpyDeviceToHost,
    cudaMemcpyKind,
    cudaMemcpyHostToDevice,
    cudaStream_t,
    cudaStreamSynchronize,
    cudaSuccess,
)

from . cimport mpi_xla_bridge


# Error handling

cpdef inline unicode py_string(const char* c_str):
    py_str = <bytes> c_str
    return py_str.decode("UTF-8")


cpdef inline unicode get_error_name(cudaError_t ierr):
    return py_string(cudaGetErrorName(ierr))


cpdef inline unicode get_error_string(cudaError_t ierr):
    return py_string(cudaGetErrorString(ierr))


# Config

cdef bint COPY_TO_HOST = False

cpdef void set_copy_to_host(bint enable):
    global COPY_TO_HOST
    COPY_TO_HOST = enable


# Memory management

cdef inline void* checked_malloc(size_t count, MPI_Comm comm) nogil:
    cdef void* mem = malloc(count)
    if not mem:
        with gil:
            message = f"Failed to allocate {count} bytes on host"

        mpi_xla_bridge.abort(0, comm, message)

    return mem


cdef inline cudaError_t checked_cuda_memcpy(void* dst, void* src, size_t count,
                                            cudaMemcpyKind kind, MPI_Comm comm) nogil:
    cdef cudaError_t ierr
    ierr = cudaMemcpy(dst, src, count, kind)

    if ierr != cudaSuccess:
        with gil:
            err_str = get_error_name(ierr)
            err_des = get_error_string(ierr)
            message = (
                f"cudaMemcpy failed with the following error:\n"
                f"\tError {ierr} {err_str}: {err_des}"
            )

        mpi_xla_bridge.abort(0, comm, message)

    return ierr


cdef inline cudaError_t checked_cuda_stream_synchronize(
    cudaStream_t stream, MPI_Comm comm
) nogil:
    cdef cudaError_t ierr
    ierr = cudaStreamSynchronize(stream)

    if ierr != cudaSuccess:
        with gil:
            err_str = get_error_name(ierr)
            err_des = get_error_string(ierr)
            message = (
                f"cudaStreamSynchronize failed with the following error:\n"
                f"\tError {ierr} {err_str}: {err_des}"
            )

        mpi_xla_bridge.abort(0, comm, message)

    return ierr


#
# GPU XLA targets
#

# Allgather

cdef struct AllgatherDescriptor:
    int sendcount
    MPI_Datatype sendtype
    int recvcount
    MPI_Datatype recvtype
    MPI_Comm comm


cpdef bytes build_allgather_descriptor(
    int sendcount, uintptr_t sendtype_handle,
    int recvcount, uintptr_t recvtype_addr,
    uintptr_t comm_handle
):
    cdef AllgatherDescriptor desc = AllgatherDescriptor(
        sendcount, <MPI_Datatype> sendtype_handle,
        recvcount, <MPI_Datatype> recvtype_addr,
        <MPI_Comm> comm_handle
    )
    return bytes((<char*> &desc)[:sizeof(AllgatherDescriptor)])


cdef void mpi_allgather_gpu(cudaStream_t stream, void** buffers,
                            const char* opaque, size_t opaque_len) nogil:
    cdef int ierr, sendtype_size, recvtype_size, comm_size
    cdef size_t sendbytes, recvbytes

    # decode inputs
    cdef void* data = buffers[0]
    cdef void* out_data = buffers[2]

    cdef void* in_buf = data
    cdef void* out_buf = out_data

    if opaque_len != sizeof(AllgatherDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef AllgatherDescriptor* desc = <AllgatherDescriptor*>(opaque)
    cdef int sendcount = desc.sendcount
    cdef MPI_Datatype sendtype = desc.sendtype
    cdef int recvcount = desc.recvcount
    cdef MPI_Datatype recvtype = desc.recvtype
    cdef MPI_Comm comm = desc.comm

    checked_cuda_stream_synchronize(stream, comm)

    if COPY_TO_HOST:
        # copy memory to host
        ierr = MPI_Type_size(sendtype, &sendtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        sendbytes = sendtype_size * sendcount
        in_buf = checked_malloc(sendbytes, comm)

        ierr = MPI_Type_size(recvtype, &recvtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        ierr = MPI_Comm_size(comm, &comm_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Comm_size")

        # recvcount is received data *per process*
        recvbytes = recvtype_size * recvcount * comm_size
        out_buf = checked_malloc(recvbytes, comm)

        checked_cuda_memcpy(in_buf, data, sendbytes, cudaMemcpyDeviceToHost, comm)

    mpi_xla_bridge.mpi_allgather(
        in_buf, sendcount, sendtype,
        out_buf, recvcount, recvtype,
        comm
    )

    if COPY_TO_HOST:
        # copy back to device
        checked_cuda_memcpy(out_data, out_buf, recvbytes, cudaMemcpyHostToDevice, comm)
        free(in_buf)
        free(out_buf)


# Allreduce

cdef struct AllreduceDescriptor:
    int nitems
    MPI_Op op
    MPI_Comm comm
    MPI_Datatype dtype


cpdef bytes build_allreduce_descriptor(int nitems, uintptr_t op_handle,
                                       uintptr_t comm_handle, uintptr_t dtype_handle):
    cdef AllreduceDescriptor desc = AllreduceDescriptor(
        nitems, <MPI_Op> op_handle, <MPI_Comm> comm_handle, <MPI_Datatype> dtype_handle
    )
    return bytes((<char*> &desc)[:sizeof(AllreduceDescriptor)])


cdef void mpi_allreduce_gpu(cudaStream_t stream, void** buffers,
                            const char* opaque, size_t opaque_len) nogil:
    cdef int ierr, dtype_size
    cdef size_t count

    # decode inputs
    cdef void* data = buffers[0]
    cdef void* out_data = buffers[2]

    cdef void* in_buf = data
    cdef void* out_buf = out_data

    if opaque_len != sizeof(AllreduceDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef AllreduceDescriptor* desc = <AllreduceDescriptor*>(opaque)
    cdef int nitems = desc.nitems
    cdef MPI_Op op = desc.op
    cdef MPI_Comm comm = desc.comm
    cdef MPI_Datatype dtype = desc.dtype

    checked_cuda_stream_synchronize(stream, comm)

    if COPY_TO_HOST:
        # copy memory to host
        ierr = MPI_Type_size(dtype, &dtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        count = dtype_size * nitems
        in_buf = checked_malloc(count, comm)
        out_buf = checked_malloc(count, comm)
        checked_cuda_memcpy(in_buf, data, count, cudaMemcpyDeviceToHost, comm)

    mpi_xla_bridge.mpi_allreduce(in_buf, out_buf, nitems, dtype, op, comm)

    if COPY_TO_HOST:
        # copy back to device
        checked_cuda_memcpy(out_data, out_buf, count, cudaMemcpyHostToDevice, comm)
        free(in_buf)
        free(out_buf)


# Alltoall

cdef struct AlltoallDescriptor:
    int sendcount
    MPI_Datatype sendtype
    int recvcount
    MPI_Datatype recvtype
    MPI_Comm comm


cpdef bytes build_alltoall_descriptor(
    int sendcount, uintptr_t sendtype_handle,
    int recvcount, uintptr_t recvtype_addr,
    uintptr_t comm_handle
):
    cdef AlltoallDescriptor desc = AlltoallDescriptor(
        sendcount, <MPI_Datatype> sendtype_handle,
        recvcount, <MPI_Datatype> recvtype_addr,
        <MPI_Comm> comm_handle
    )
    return bytes((<char*> &desc)[:sizeof(AlltoallDescriptor)])


cdef void mpi_alltoall_gpu(cudaStream_t stream, void** buffers,
                           const char* opaque, size_t opaque_len) nogil:
    cdef int ierr, sendtype_size, recvtype_size, comm_size
    cdef size_t sendbytes, recvbytes

    # decode inputs
    cdef void* data = buffers[0]
    cdef void* out_data = buffers[2]

    cdef void* in_buf = data
    cdef void* out_buf = out_data

    if opaque_len != sizeof(AlltoallDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef AlltoallDescriptor* desc = <AlltoallDescriptor*>(opaque)
    cdef int sendcount = desc.sendcount
    cdef MPI_Datatype sendtype = desc.sendtype
    cdef int recvcount = desc.recvcount
    cdef MPI_Datatype recvtype = desc.recvtype
    cdef MPI_Comm comm = desc.comm

    checked_cuda_stream_synchronize(stream, comm)

    if COPY_TO_HOST:
        # copy memory to host

        # counts are *per process*, so multiply with comm_size
        ierr = MPI_Comm_size(comm, &comm_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Comm_size")

        ierr = MPI_Type_size(sendtype, &sendtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        sendbytes = sendtype_size * sendcount * comm_size
        in_buf = checked_malloc(sendbytes, comm)

        ierr = MPI_Type_size(recvtype, &recvtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        recvbytes = recvtype_size * recvcount * comm_size
        out_buf = checked_malloc(recvbytes, comm)

        checked_cuda_memcpy(in_buf, data, sendbytes, cudaMemcpyDeviceToHost, comm)

    mpi_xla_bridge.mpi_alltoall(
        in_buf, sendcount, sendtype, out_buf, recvcount, recvtype, comm
    )

    if COPY_TO_HOST:
        # copy back to device
        checked_cuda_memcpy(out_data, out_buf, recvbytes, cudaMemcpyHostToDevice, comm)
        free(in_buf)
        free(out_buf)


# Barrier

cdef struct BarrierDescriptor:
    MPI_Comm comm


cpdef bytes build_barrier_descriptor(uintptr_t comm_handle):
    cdef BarrierDescriptor desc = BarrierDescriptor(<MPI_Comm> comm_handle)
    return bytes((<char*> &desc)[:sizeof(BarrierDescriptor)])


cdef void mpi_barrier_gpu(cudaStream_t stream, void** buffers,
                          const char* opaque, size_t opaque_len) nogil:
    if opaque_len != sizeof(BarrierDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef BarrierDescriptor* desc = <BarrierDescriptor*>(opaque)
    cdef MPI_Comm comm = desc.comm

    mpi_xla_bridge.mpi_barrier(comm)


# Bcast

cdef struct BcastDescriptor:
    int nitems
    int root
    MPI_Comm comm
    MPI_Datatype dtype


cpdef bytes build_bcast_descriptor(int nitems, int root, uintptr_t comm_handle,
                                   uintptr_t dtype_handle):
    cdef BcastDescriptor desc = BcastDescriptor(
        nitems, root, <MPI_Comm> comm_handle, <MPI_Datatype> dtype_handle
    )
    return bytes((<char*> &desc)[:sizeof(BcastDescriptor)])


cdef void mpi_bcast_gpu(cudaStream_t stream, void** buffers,
                        const char* opaque, size_t opaque_len) nogil:
    cdef int ierr, dtype_size, rank
    cdef size_t count

    # decode inputs
    cdef void* data = buffers[0]
    cdef void* out_data = buffers[2]

    cdef void* buf = out_data

    if opaque_len != sizeof(BcastDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef BcastDescriptor* desc = <BcastDescriptor*>(opaque)
    cdef int nitems = desc.nitems
    cdef int root = desc.root
    cdef MPI_Comm comm = desc.comm
    cdef MPI_Datatype dtype = desc.dtype

    ierr = MPI_Type_size(dtype, &dtype_size)
    mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")
    ierr = MPI_Comm_rank(comm, &rank)
    mpi_xla_bridge.abort_on_error(ierr, comm, u"Comm_rank")

    checked_cuda_stream_synchronize(stream, comm)

    if COPY_TO_HOST:
        # copy memory to host
        count = dtype_size * nitems
        buf = checked_malloc(count, comm)
        if rank == root:
            checked_cuda_memcpy(buf, data, count, cudaMemcpyDeviceToHost, comm)
    else:
        if rank == root:
            buf = data

    mpi_xla_bridge.mpi_bcast(buf, nitems, dtype, root, comm)

    if COPY_TO_HOST:
        if rank != root:
            # copy back to device
            checked_cuda_memcpy(out_data, buf, count, cudaMemcpyHostToDevice, comm)
        free(buf)


# Gather

cdef struct GatherDescriptor:
    int sendcount
    MPI_Datatype sendtype
    int recvcount
    MPI_Datatype recvtype
    int root
    MPI_Comm comm


cpdef bytes build_gather_descriptor(
    int sendcount, uintptr_t sendtype_handle,
    int recvcount, uintptr_t recvtype_addr,
    int root, uintptr_t comm_handle
):
    cdef GatherDescriptor desc = GatherDescriptor(
        sendcount, <MPI_Datatype> sendtype_handle,
        recvcount, <MPI_Datatype> recvtype_addr,
        root, <MPI_Comm> comm_handle
    )
    return bytes((<char*> &desc)[:sizeof(GatherDescriptor)])


cdef void mpi_gather_gpu(cudaStream_t stream, void** buffers,
                         const char* opaque, size_t opaque_len) nogil:
    cdef int ierr, sendtype_size, recvtype_size, rank, size
    cdef size_t sendbytes, recvbytes

    # decode inputs
    cdef void* data = buffers[0]
    cdef void* out_data = buffers[2]

    cdef void* in_buf = data
    cdef void* out_buf = out_data

    if opaque_len != sizeof(GatherDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef GatherDescriptor* desc = <GatherDescriptor*>(opaque)
    cdef int sendcount = desc.sendcount
    cdef MPI_Datatype sendtype = desc.sendtype
    cdef int recvcount = desc.recvcount
    cdef MPI_Datatype recvtype = desc.recvtype
    cdef int root = desc.root
    cdef MPI_Comm comm = desc.comm

    checked_cuda_stream_synchronize(stream, comm)

    if COPY_TO_HOST:
        # copy memory to host
        ierr = MPI_Type_size(sendtype, &sendtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        sendbytes = sendtype_size * sendcount
        in_buf = checked_malloc(sendbytes, comm)

        ierr = MPI_Type_size(recvtype, &recvtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        ierr = MPI_Comm_rank(comm, &rank)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Comm_rank")

        ierr = MPI_Comm_size(comm, &size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Comm_size")

        recvbytes = recvtype_size * recvcount
        if rank == root:
            # output shape is larger on root
            recvbytes *= size

        out_buf = checked_malloc(recvbytes, comm)

        checked_cuda_memcpy(in_buf, data, sendbytes, cudaMemcpyDeviceToHost, comm)

    mpi_xla_bridge.mpi_gather(
        in_buf, sendcount, sendtype, out_buf, recvcount, recvtype, root, comm
    )

    if COPY_TO_HOST:
        if rank == root:
            # copy back to device
            checked_cuda_memcpy(
                out_data, out_buf, recvbytes, cudaMemcpyHostToDevice, comm
            )

        free(in_buf)
        free(out_buf)


# Recv

cdef struct RecvDescriptor:
    int nitems
    int source
    int tag
    MPI_Comm comm
    MPI_Datatype dtype
    MPI_Status* status


cpdef bytes build_recv_descriptor(int nitems, int dest, int tag, uintptr_t comm_handle,
                                  uintptr_t dtype_handle, uintptr_t status_addr):
    cdef RecvDescriptor desc = RecvDescriptor(
        nitems, dest, tag,
        <MPI_Comm> comm_handle,
        <MPI_Datatype> dtype_handle,
        <MPI_Status*> status_addr
    )
    return bytes((<char*> &desc)[:sizeof(RecvDescriptor)])


cdef void mpi_recv_gpu(cudaStream_t stream, void** buffers,
                       const char* opaque, size_t opaque_len) nogil:
    cdef int ierr, dtype_size
    cdef size_t count

    # decode inputs
    cdef void* out_buf = buffers[1]

    cdef void* recvbuf = out_buf

    if opaque_len != sizeof(RecvDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef RecvDescriptor* desc = <RecvDescriptor*>(opaque)
    cdef int nitems = desc.nitems
    cdef int source = desc.source
    cdef int tag = desc.tag
    cdef MPI_Comm comm = desc.comm
    cdef MPI_Datatype dtype = desc.dtype
    cdef MPI_Status* status = desc.status

    checked_cuda_stream_synchronize(stream, comm)

    if COPY_TO_HOST:
        # copy memory to host
        ierr = MPI_Type_size(dtype, &dtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        count = dtype_size * nitems
        recvbuf = checked_malloc(count, comm)

    mpi_xla_bridge.mpi_recv(recvbuf, nitems, dtype, source, tag, comm, status)

    if COPY_TO_HOST:
        checked_cuda_memcpy(out_buf, recvbuf, count, cudaMemcpyHostToDevice, comm)
        free(recvbuf)


# Reduce

cdef struct ReduceDescriptor:
    int nitems
    MPI_Op op
    int root
    MPI_Comm comm
    MPI_Datatype dtype


cpdef bytes build_reduce_descriptor(int nitems, uintptr_t op_handle, int root,
                                    uintptr_t comm_handle, uintptr_t dtype_handle):
    cdef ReduceDescriptor desc = ReduceDescriptor(
        nitems, <MPI_Op> op_handle, root, <MPI_Comm> comm_handle,
        <MPI_Datatype> dtype_handle
    )
    return bytes((<char*> &desc)[:sizeof(ReduceDescriptor)])


cdef void mpi_reduce_gpu(cudaStream_t stream, void** buffers,
                         const char* opaque, size_t opaque_len) nogil:
    cdef int ierr, dtype_size, rank
    cdef size_t count

    # decode inputs
    cdef void* data = buffers[0]
    cdef void* out_data = buffers[2]

    cdef void* in_buf = data
    cdef void* out_buf = out_data

    if opaque_len != sizeof(ReduceDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef ReduceDescriptor* desc = <ReduceDescriptor*>(opaque)
    cdef int nitems = desc.nitems
    cdef MPI_Op op = desc.op
    cdef int root = desc.root
    cdef MPI_Comm comm = desc.comm
    cdef MPI_Datatype dtype = desc.dtype

    checked_cuda_stream_synchronize(stream, comm)

    if COPY_TO_HOST:
        # copy memory to host
        ierr = MPI_Type_size(dtype, &dtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        count = dtype_size * nitems
        in_buf = checked_malloc(count, comm)
        out_buf = checked_malloc(count, comm)
        checked_cuda_memcpy(in_buf, data, count, cudaMemcpyDeviceToHost, comm)

    mpi_xla_bridge.mpi_reduce(in_buf, out_buf, nitems, dtype, op, root, comm)

    if COPY_TO_HOST:
        ierr = MPI_Comm_rank(comm, &rank)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Comm_rank")

        if rank == root:
            # copy back to device
            checked_cuda_memcpy(out_data, out_buf, count, cudaMemcpyHostToDevice, comm)

        free(in_buf)
        free(out_buf)


# Scan

cdef struct ScanDescriptor:
    int nitems
    MPI_Op op
    MPI_Comm comm
    MPI_Datatype dtype


cpdef bytes build_scan_descriptor(int nitems, uintptr_t op_handle,
                                  uintptr_t comm_handle, uintptr_t dtype_handle):
    cdef ScanDescriptor desc = ScanDescriptor(
        nitems, <MPI_Op> op_handle, <MPI_Comm> comm_handle, <MPI_Datatype> dtype_handle
    )
    return bytes((<char*> &desc)[:sizeof(ScanDescriptor)])


cdef void mpi_scan_gpu(cudaStream_t stream, void** buffers,
                       const char* opaque, size_t opaque_len) nogil:
    cdef int ierr, dtype_size
    cdef size_t count

    # decode inputs
    cdef void* data = buffers[0]
    cdef void* out_data = buffers[2]

    cdef void* in_buf = data
    cdef void* out_buf = out_data

    if opaque_len != sizeof(ScanDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef ScanDescriptor* desc = <ScanDescriptor*>(opaque)
    cdef int nitems = desc.nitems
    cdef MPI_Op op = desc.op
    cdef MPI_Comm comm = desc.comm
    cdef MPI_Datatype dtype = desc.dtype

    checked_cuda_stream_synchronize(stream, comm)

    if COPY_TO_HOST:
        # copy memory to host
        ierr = MPI_Type_size(dtype, &dtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        count = dtype_size * nitems
        in_buf = checked_malloc(count, comm)
        out_buf = checked_malloc(count, comm)
        checked_cuda_memcpy(in_buf, data, count, cudaMemcpyDeviceToHost, comm)

    mpi_xla_bridge.mpi_scan(in_buf, out_buf, nitems, dtype, op, comm)

    if COPY_TO_HOST:
        # copy back to device
        checked_cuda_memcpy(out_data, out_buf, count, cudaMemcpyHostToDevice, comm)
        free(in_buf)
        free(out_buf)


# Scatter

cdef struct ScatterDescriptor:
    int sendcount
    MPI_Datatype sendtype
    int recvcount
    MPI_Datatype recvtype
    int root
    MPI_Comm comm


cpdef bytes build_scatter_descriptor(
    int sendcount, uintptr_t sendtype_handle,
    int recvcount, uintptr_t recvtype_addr,
    int root, uintptr_t comm_handle
):
    cdef ScatterDescriptor desc = ScatterDescriptor(
        sendcount, <MPI_Datatype> sendtype_handle,
        recvcount, <MPI_Datatype> recvtype_addr,
        root, <MPI_Comm> comm_handle
    )
    return bytes((<char*> &desc)[:sizeof(ScatterDescriptor)])


cdef void mpi_scatter_gpu(cudaStream_t stream, void** buffers,
                          const char* opaque, size_t opaque_len) nogil:
    cdef int ierr, sendtype_size, recvtype_size, rank, size
    cdef size_t sendbytes, recvbytes

    # decode inputs
    cdef void* data = buffers[0]
    cdef void* out_data = buffers[2]

    cdef void* in_buf = data
    cdef void* out_buf = out_data

    if opaque_len != sizeof(ScatterDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef ScatterDescriptor* desc = <ScatterDescriptor*>(opaque)
    cdef int sendcount = desc.sendcount
    cdef MPI_Datatype sendtype = desc.sendtype
    cdef int recvcount = desc.recvcount
    cdef MPI_Datatype recvtype = desc.recvtype
    cdef int root = desc.root
    cdef MPI_Comm comm = desc.comm

    checked_cuda_stream_synchronize(stream, comm)

    if COPY_TO_HOST:
        # copy memory to host
        ierr = MPI_Comm_rank(comm, &rank)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Comm_rank")

        ierr = MPI_Comm_size(comm, &size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Comm_size")

        ierr = MPI_Type_size(sendtype, &sendtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        sendbytes = sendtype_size * sendcount
        if rank == root:
            # input is larger on root
            sendbytes *= size

        in_buf = checked_malloc(sendbytes, comm)

        ierr = MPI_Type_size(recvtype, &recvtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        recvbytes = recvtype_size * recvcount
        out_buf = checked_malloc(recvbytes, comm)

        checked_cuda_memcpy(in_buf, data, sendbytes, cudaMemcpyDeviceToHost, comm)

    mpi_xla_bridge.mpi_scatter(
        in_buf, sendcount, sendtype, out_buf, recvcount, recvtype, root, comm
    )

    if COPY_TO_HOST:
        # copy back to device
        checked_cuda_memcpy(out_data, out_buf, recvbytes, cudaMemcpyHostToDevice, comm)

        free(in_buf)
        free(out_buf)


# Send

cdef struct SendDescriptor:
    int nitems
    int dest
    int tag
    MPI_Comm comm
    MPI_Datatype dtype


cpdef bytes build_send_descriptor(int nitems, int dest, int tag, uintptr_t comm_handle,
                                  uintptr_t dtype_handle):
    cdef SendDescriptor desc = SendDescriptor(
        nitems, dest, tag, <MPI_Comm> comm_handle, <MPI_Datatype> dtype_handle
    )
    return bytes((<char*> &desc)[:sizeof(SendDescriptor)])


cdef void mpi_send_gpu(cudaStream_t stream, void** buffers,
                       const char* opaque, size_t opaque_len) nogil:
    cdef int ierr, dtype_size
    cdef size_t count

    # decode inputs
    cdef void* data = buffers[0]

    cdef void* sendbuf = data

    if opaque_len != sizeof(SendDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef SendDescriptor* desc = <SendDescriptor*>(opaque)
    cdef int nitems = desc.nitems
    cdef int dest = desc.dest
    cdef int tag = desc.tag
    cdef MPI_Comm comm = desc.comm
    cdef MPI_Datatype dtype = desc.dtype

    checked_cuda_stream_synchronize(stream, comm)

    if COPY_TO_HOST:
        # copy memory to host
        ierr = MPI_Type_size(dtype, &dtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        count = dtype_size * nitems
        sendbuf = checked_malloc(count, comm)
        checked_cuda_memcpy(sendbuf, data, count, cudaMemcpyDeviceToHost, comm)

    mpi_xla_bridge.mpi_send(sendbuf, nitems, dtype, dest, tag, comm)

    if COPY_TO_HOST:
        free(sendbuf)


# Sendrecv

cdef struct SendrecvDescriptor:
    int sendcount
    int dest
    int sendtag
    MPI_Datatype sendtype
    int recvcount
    int source
    int recvtag
    MPI_Datatype recvtype
    MPI_Comm comm
    MPI_Status* status


cpdef bytes build_sendrecv_descriptor(
    int sendcount, int dest, int sendtag, uintptr_t sendtype_handle,
    int recvcount, int source, int recvtag, uintptr_t recvtype_addr,
    uintptr_t comm_handle, uintptr_t status_addr
):
    cdef SendrecvDescriptor desc = SendrecvDescriptor(
        sendcount, dest, sendtag, <MPI_Datatype> sendtype_handle,
        recvcount, source, recvtag, <MPI_Datatype> recvtype_addr,
        <MPI_Comm> comm_handle, <MPI_Status*> status_addr
    )
    return bytes((<char*> &desc)[:sizeof(SendrecvDescriptor)])


cdef void mpi_sendrecv_gpu(cudaStream_t stream, void** buffers,
                           const char* opaque, size_t opaque_len) nogil:
    cdef int ierr, send_dtype_size, recv_dtype_size
    cdef size_t bytes_send, bytes_recv

    # decode inputs
    cdef void* in_buf = buffers[0]
    cdef void* out_buf = buffers[2]

    cdef void* sendbuf = in_buf
    cdef void* recvbuf = out_buf

    if opaque_len != sizeof(SendrecvDescriptor):
        with gil:
            raise RuntimeError("got wrong size of opaque argument")

    cdef SendrecvDescriptor* desc = <SendrecvDescriptor*>(opaque)
    cdef int sendcount = desc.sendcount
    cdef int dest = desc.dest
    cdef int sendtag = desc.sendtag
    cdef MPI_Datatype sendtype = desc.sendtype
    cdef int recvcount = desc.recvcount
    cdef int source = desc.source
    cdef int recvtag = desc.recvtag
    cdef MPI_Datatype recvtype = desc.recvtype
    cdef MPI_Comm comm = desc.comm
    cdef MPI_Status* status = desc.status

    checked_cuda_stream_synchronize(stream, comm)

    if COPY_TO_HOST:
        # copy memory to host
        ierr = MPI_Type_size(sendtype, &send_dtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        ierr = MPI_Type_size(recvtype, &recv_dtype_size)
        mpi_xla_bridge.abort_on_error(ierr, comm, u"Type_size")

        bytes_send = send_dtype_size * sendcount
        bytes_recv = recv_dtype_size * recvcount
        sendbuf = checked_malloc(bytes_send, comm)
        recvbuf = checked_malloc(bytes_recv, comm)
        checked_cuda_memcpy(sendbuf, in_buf, bytes_send, cudaMemcpyDeviceToHost, comm)

    mpi_xla_bridge.mpi_sendrecv(
        sendbuf, sendcount, sendtype, dest, sendtag,
        recvbuf, recvcount, recvtype, source, recvtag,
        comm, status
    )

    if COPY_TO_HOST:
        checked_cuda_memcpy(out_buf, recvbuf, bytes_recv, cudaMemcpyHostToDevice, comm)
        free(recvbuf)


gpu_custom_call_targets = {}

cdef register_custom_call_target(fn_name, void* fn):
    cdef const char* name = "xla._CUSTOM_CALL_TARGET"
    gpu_custom_call_targets[fn_name] = PyCapsule_New(fn, name, NULL)


register_custom_call_target(b"mpi_allgather", <void*>(mpi_allgather_gpu))
register_custom_call_target(b"mpi_allreduce", <void*>(mpi_allreduce_gpu))
register_custom_call_target(b"mpi_alltoall", <void*>(mpi_alltoall_gpu))
register_custom_call_target(b"mpi_barrier", <void*>(mpi_barrier_gpu))
register_custom_call_target(b"mpi_bcast", <void*>(mpi_bcast_gpu))
register_custom_call_target(b"mpi_gather", <void*>(mpi_gather_gpu))
register_custom_call_target(b"mpi_recv", <void*>(mpi_recv_gpu))
register_custom_call_target(b"mpi_reduce", <void*>(mpi_reduce_gpu))
register_custom_call_target(b"mpi_scan", <void*>(mpi_scan_gpu))
register_custom_call_target(b"mpi_scatter", <void*>(mpi_scatter_gpu))
register_custom_call_target(b"mpi_send", <void*>(mpi_send_gpu))
register_custom_call_target(b"mpi_sendrecv", <void*>(mpi_sendrecv_gpu))



import os
from jax.lib import xla_client

from . import mpi_xla_bridge
from . import mpi_xla_bridge_cpu

try:
    from . import mpi_xla_bridge_gpu
except ImportError:
    HAS_GPU_EXT = False
else:
    HAS_GPU_EXT = True


# setup logging


def _is_truthy(str_val):
    return str_val.lower() in ("true", "1", "on")


mpi_xla_bridge.set_logging(_is_truthy(os.getenv("MPI4JAX_DEBUG", "")))


# register custom call targets
for name, fn in mpi_xla_bridge_cpu.cpu_custom_call_targets.items():
    xla_client.register_custom_call_target(name, fn, platform="cpu")

if HAS_GPU_EXT:
    for name, fn in mpi_xla_bridge_gpu.gpu_custom_call_targets.items():
        xla_client.register_custom_call_target(name, fn, platform="CUDA")



from cpython.pycapsule cimport PyCapsule_New

from libc.stdint cimport uintptr_t
from libc.string cimport memcpy

from mpi4py.libmpi cimport (
    MPI_Comm,
    MPI_Comm_rank,
    MPI_Datatype,
    MPI_Op,
    MPI_Status,
)

from . cimport mpi_xla_bridge


#
# CPU XLA targets
#
cdef void mpi_barrier_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef MPI_Comm comm = <MPI_Comm>((<uintptr_t*>(data_ptr[0]))[0])
    mpi_xla_bridge.mpi_barrier(comm)


cdef void mpi_allgather_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int sendcount = (<int*>(data_ptr[0]))[0]
    cdef void* sendbuf = data_ptr[1]
    cdef MPI_Datatype sendtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[2]))[0])
    cdef int recvcount = (<int*>(data_ptr[3]))[0]
    cdef MPI_Datatype recvtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[4]))[0])
    cdef MPI_Comm comm = <MPI_Comm>((<uintptr_t*>(data_ptr[5]))[0])

    cdef void* recvbuf = out_ptr[0]
    mpi_xla_bridge.mpi_allgather(
        sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, comm
    )


cdef void mpi_allreduce_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int nitems = (<int*>(data_ptr[0]))[0]
    cdef void* sendbuf = data_ptr[1]
    cdef MPI_Op op = <MPI_Op>((<uintptr_t*>(data_ptr[2]))[0])
    cdef MPI_Comm comm = <MPI_Comm>((<uintptr_t*>(data_ptr[3]))[0])
    cdef MPI_Datatype dtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[4]))[0])

    cdef void* recvbuf = out_ptr[0]
    mpi_xla_bridge.mpi_allreduce(
        sendbuf, recvbuf, nitems, dtype, op, comm
    )


cdef void mpi_alltoall_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int sendcount = (<int*>(data_ptr[0]))[0]
    cdef void* sendbuf = data_ptr[1]
    cdef MPI_Datatype sendtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[2]))[0])
    cdef int recvcount = (<int*>(data_ptr[3]))[0]
    cdef MPI_Datatype recvtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[4]))[0])
    cdef MPI_Comm comm = <MPI_Comm>((<uintptr_t*>(data_ptr[5]))[0])

    cdef void* recvbuf = out_ptr[0]
    mpi_xla_bridge.mpi_alltoall(
        sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, comm
    )


cdef void mpi_bcast_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int nitems = (<int*>(data_ptr[0]))[0]
    cdef void* sendbuf = data_ptr[1]
    cdef int root = (<int*>(data_ptr[2]))[0]
    cdef MPI_Comm comm = <MPI_Comm>((<uintptr_t*>(data_ptr[3]))[0])
    cdef MPI_Datatype dtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[4]))[0])

    cdef void* recvbuf = out_ptr[0]

    cdef int ierr, rank
    cdef void* buf
    ierr = MPI_Comm_rank(comm, &rank)
    mpi_xla_bridge.abort_on_error(ierr, comm, u"Comm_rank")

    if rank == root:
        buf = sendbuf
    else:
        buf = recvbuf

    mpi_xla_bridge.mpi_bcast(buf, nitems, dtype, root, comm)


cdef void mpi_gather_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int sendcount = (<int*>(data_ptr[0]))[0]
    cdef void* sendbuf = data_ptr[1]
    cdef MPI_Datatype sendtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[2]))[0])
    cdef int recvcount = (<int*>(data_ptr[3]))[0]
    cdef MPI_Datatype recvtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[4]))[0])
    cdef int root = (<int*>(data_ptr[5]))[0]
    cdef MPI_Comm comm = <MPI_Comm>((<uintptr_t*>(data_ptr[6]))[0])

    cdef void* recvbuf = out_ptr[0]

    mpi_xla_bridge.mpi_gather(
        sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm
    )


cdef void mpi_recv_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int nitems = (<int*>(data_ptr[0]))[0]
    cdef int source = (<int*>(data_ptr[1]))[0]
    cdef int tag = (<int*>(data_ptr[2]))[0]
    cdef MPI_Comm comm = <MPI_Comm>((<uintptr_t*>(data_ptr[3]))[0])
    cdef MPI_Datatype dtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[4]))[0])
    cdef MPI_Status* status = <MPI_Status*>((<uintptr_t*>(data_ptr[5]))[0])

    cdef void* recvbuf = out_ptr[0]
    mpi_xla_bridge.mpi_recv(recvbuf, nitems, dtype, source, tag, comm, status)


cdef void mpi_reduce_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int nitems = (<int*>(data_ptr[0]))[0]
    cdef void* sendbuf = data_ptr[1]
    cdef MPI_Op op = <MPI_Op>((<uintptr_t*>(data_ptr[2]))[0])
    cdef int root = (<int*>(data_ptr[3]))[0]
    cdef MPI_Comm comm = <MPI_Comm>((<uintptr_t*>(data_ptr[4]))[0])
    cdef MPI_Datatype dtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[5]))[0])

    cdef void* recvbuf = out_ptr[0]
    mpi_xla_bridge.mpi_reduce(
        sendbuf, recvbuf, nitems, dtype, op, root, comm
    )


cdef void mpi_scan_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int nitems = (<int*>(data_ptr[0]))[0]
    cdef void* sendbuf = data_ptr[1]
    cdef MPI_Op op = <MPI_Op>((<uintptr_t*>(data_ptr[2]))[0])
    cdef MPI_Comm comm = <MPI_Comm>((<uintptr_t*>(data_ptr[3]))[0])
    cdef MPI_Datatype dtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[4]))[0])

    cdef void* recvbuf = out_ptr[0]
    mpi_xla_bridge.mpi_scan(
        sendbuf, recvbuf, nitems, dtype, op, comm
    )


cdef void mpi_scatter_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int sendcount = (<int*>(data_ptr[0]))[0]
    cdef void* sendbuf = data_ptr[1]
    cdef MPI_Datatype sendtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[2]))[0])
    cdef int recvcount = (<int*>(data_ptr[3]))[0]
    cdef MPI_Datatype recvtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[4]))[0])
    cdef int root = (<int*>(data_ptr[5]))[0]
    cdef MPI_Comm comm = <MPI_Comm>((<uintptr_t*>(data_ptr[6]))[0])

    cdef void* recvbuf = out_ptr[0]
    mpi_xla_bridge.mpi_scatter(
        sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm
    )


cdef void mpi_send_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int nitems = (<int*>(data_ptr[0]))[0]
    cdef void* sendbuf = data_ptr[1]
    cdef int destination = (<int*>(data_ptr[2]))[0]
    cdef int tag = (<int*>(data_ptr[3]))[0]
    cdef MPI_Comm comm = <MPI_Comm>((<uintptr_t*>(data_ptr[4]))[0])
    cdef MPI_Datatype dtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[5]))[0])

    mpi_xla_bridge.mpi_send(sendbuf, nitems, dtype, destination, tag, comm)


cdef void mpi_sendrecv_cpu(void** out_ptr, void** data_ptr) nogil:
    cdef int sendcount = (<int*>(data_ptr[0]))[0]
    cdef void* sendbuf = data_ptr[1]
    cdef int dest = (<int*>(data_ptr[2]))[0]
    cdef int sendtag = (<int*>(data_ptr[3]))[0]
    cdef MPI_Datatype sendtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[4]))[0])

    cdef int recvcount = (<int*>(data_ptr[5]))[0]
    cdef int source = (<int*>(data_ptr[6]))[0]
    cdef int recvtag = (<int*>(data_ptr[7]))[0]
    cdef MPI_Datatype recvtype = <MPI_Datatype>((<uintptr_t*>(data_ptr[8]))[0])

    cdef MPI_Comm comm = <MPI_Comm>((<uintptr_t*>(data_ptr[9]))[0])
    cdef MPI_Status* status = <MPI_Status*>((<uintptr_t*>(data_ptr[10]))[0])

    cdef void* recvbuf = out_ptr[0]
    mpi_xla_bridge.mpi_sendrecv(
        sendbuf, sendcount, sendtype, dest, sendtag,
        recvbuf, recvcount, recvtype, source, recvtag,
        comm, status
    )


cpu_custom_call_targets = {}

cdef register_custom_call_target(fn_name, void* fn):
    cdef const char* name = "xla._CUSTOM_CALL_TARGET"
    cpu_custom_call_targets[fn_name] = PyCapsule_New(fn, name, NULL)

register_custom_call_target(b"mpi_allgather", <void*>(mpi_allgather_cpu))
register_custom_call_target(b"mpi_allreduce", <void*>(mpi_allreduce_cpu))
register_custom_call_target(b"mpi_alltoall", <void*>(mpi_alltoall_cpu))
register_custom_call_target(b"mpi_barrier", <void*>(mpi_barrier_cpu))
register_custom_call_target(b"mpi_bcast", <void*>(mpi_bcast_cpu))
register_custom_call_target(b"mpi_gather", <void*>(mpi_gather_cpu))
register_custom_call_target(b"mpi_recv", <void*>(mpi_recv_cpu))
register_custom_call_target(b"mpi_reduce", <void*>(mpi_reduce_cpu))
register_custom_call_target(b"mpi_scan", <void*>(mpi_scan_cpu))
register_custom_call_target(b"mpi_scatter", <void*>(mpi_scatter_cpu))
register_custom_call_target(b"mpi_send", <void*>(mpi_send_cpu))
register_custom_call_target(b"mpi_sendrecv", <void*>(mpi_sendrecv_cpu))



from mpi4py import MPI

import jax
import jax.numpy as jnp

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()


def test_bcast():
    from mpi4jax import bcast

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    if rank != 0:
        _arr = _arr * 0

    res, token = bcast(_arr, root=0)
    assert jnp.array_equal(res, arr)
    if rank == 0:
        assert jnp.array_equal(_arr, arr)


def test_bcast_jit():
    from mpi4jax import bcast

    arr = jnp.ones((3, 2))
    _arr = arr.copy()

    if rank != 0:
        _arr = _arr * 0

    res = jax.jit(lambda x: bcast(x, root=0)[0])(arr)
    assert jnp.array_equal(res, arr)
    if rank == 0:
        assert jnp.array_equal(_arr, arr)


def test_bcast_scalar():
    from mpi4jax import bcast

    arr = 1
    _arr = 1

    if rank != 0:
        _arr = _arr * 0

    res, token = bcast(_arr, root=0)
    assert jnp.array_equal(res, arr)
    if rank == 0:
        assert jnp.array_equal(_arr, arr)


def test_bcast_scalar_jit():
    from mpi4jax import bcast

    arr = 1
    _arr = 1

    if rank != 0:
        _arr = _arr * 0

    res = jax.jit(lambda x: bcast(x, root=0)[0])(_arr)
    assert jnp.array_equal(res, arr)
    if rank == 0:
        assert jnp.array_equal(_arr, arr)



# make sure to import mpi4py before anything else
# this calls MPI_Init and registers mpi4py's atexit handler
from mpi4py import MPI

# check version of JAX and jaxlib
from . import jax_compat

jax_compat.check_jax_version()  # noqa: F401

# this registers our custom XLA functions
from . import xla_bridge  # noqa: E402

# register atexit handler to flush JAX buffers
import atexit  # noqa: E402
from . import flush  # noqa: E402

atexit.register(flush.flush)

# import public API
from .collective_ops.allgather import allgather  # noqa: F401, E402
from .collective_ops.allreduce import allreduce  # noqa: F401, E402
from .collective_ops.alltoall import alltoall  # noqa: F401, E402
from .collective_ops.barrier import barrier  # noqa: F401, E402
from .collective_ops.bcast import bcast  # noqa: F401, E402
from .collective_ops.gather import gather  # noqa: F401, E402
from .collective_ops.recv import recv  # noqa: F401, E402
from .collective_ops.reduce import reduce  # noqa: F401, E402
from .collective_ops.scan import scan  # noqa: F401, E402
from .collective_ops.scatter import scatter  # noqa: F401, E402
from .collective_ops.send import send  # noqa: F401, E402
from .collective_ops.sendrecv import sendrecv  # noqa: F401, E402

from .utils import has_cuda_support  # noqa: F401, E402

# sanitize namespace
del jax_compat, xla_bridge, MPI, atexit, flush



import numpy as _np
from mpi4py import MPI as _MPI

from jax import abstract_arrays, core
from jax.core import Primitive, Tracer, Token
from jax.interpreters import ad, batching
from jax.lax import create_token

from jax.interpreters import mlir
import jaxlib.mlir.ir as ir

from ..utils import (
    HashableMPIType,
    default_primitive_impl,
    to_dtype_handle,
    to_mpi_handle,
    to_mpi_ptr,
    unpack_hashable,
    wrap_as_hashable,
    as_mhlo_constant,
    get_default_layouts,
    effect,
)
from ..jax_compat import hlo_custom_call, token_type
from ..decorators import translation_rule_cpu, translation_rule_gpu
from ..validation import enforce_types
from ..comm import get_default_comm


# The Jax primitive
mpi_sendrecv_p = Primitive("sendrecv_mpi")  # Create the primitive
mpi_sendrecv_impl = default_primitive_impl(mpi_sendrecv_p)


# This function applies the primitive to an AST
@enforce_types(
    source=_np.integer,
    dest=_np.integer,
    sendtag=_np.integer,
    recvtag=_np.integer,
    comm=(type(None), _MPI.Intracomm, HashableMPIType),
    status=(type(None), _MPI.Status, HashableMPIType),
    token=(type(None), Token, Tracer),
)
def sendrecv(
    sendbuf,
    recvbuf,
    source,
    dest,
    *,
    sendtag=0,
    recvtag=_MPI.ANY_TAG,
    comm=None,
    status=None,
    token=None,
):
    """Perform a sendrecv operation.

    .. warning::

        Unlike mpi4py's sendrecv, this returns a *new* array with the received data.

    Arguments:
        sendbuf: Array or scalar input to send.
        recvbuf: Array or scalar input with the correct shape and dtype. This can
           contain arbitrary data and will not be overwritten.
        source (int): Rank of the source MPI process.
        dest (int): Rank of the destination MPI process.
        sendtag (int): Tag of this message for sending.
        recvtag (int): Tag of this message for receiving.
        comm (mpi4py.MPI.Comm): The MPI communicator to use (defaults to
            a clone of :obj:`COMM_WORLD`).
        status (mpi4py.MPI.Status): Status object, can be used for introspection.
        token (Token): XLA token to use to ensure correct execution order.
            If not given, a new token is generated.

    Returns:
        Tuple[DeviceArray, Token]:
            - Received data.
            - A new, modified token, that depends on this operation.

    """
    if token is None:
        token = create_token(sendbuf)

    if comm is None:
        comm = get_default_comm()

    comm = wrap_as_hashable(comm)

    if status is not None:
        status = wrap_as_hashable(status)

    return tuple(
        mpi_sendrecv_p.bind(
            sendbuf,
            recvbuf,
            token,
            source=source,
            dest=dest,
            sendtag=sendtag,
            recvtag=recvtag,
            comm=comm,
            status=status,
            _must_transpose=False,
        )
    )


# This function compiles the operation
@translation_rule_cpu
def mpi_sendrecv_xla_encode_cpu(
    ctx,
    sendbuf,
    recvbuf,
    token,
    source,
    dest,
    sendtag,
    recvtag,
    comm,
    status,
    _must_transpose=False,
):
    from ..xla_bridge.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR

    # when performing forward diff, the gradient will follow the sent message.
    # so if you do a sendrecv from rank 0 to 1, the gradient wrt the inputs of rank 0
    # will end up in rank 1.\
    # it's maybe possible to fix this by, at the end of the calculation, bringing back
    # the gradient to the correct rank, but that would require some study.
    if _must_transpose:
        raise RuntimeError(
            "sendrecv cannot be used with forward-mode (vjp) autodiff, because "
            "the gradient might be located on a different mpi rank than the "
            "desired one. Use reverse-mode (jvp) differentiation instead."
        )

    comm = unpack_hashable(comm)
    status = unpack_hashable(status)

    send_aval, recv_aval, *_ = ctx.avals_in
    send_nptype = send_aval.dtype
    recv_nptype = recv_aval.dtype

    send_type = ir.RankedTensorType(sendbuf.type)
    send_dims = send_type.shape

    recv_type = ir.RankedTensorType(recvbuf.type)
    recv_dtype = recv_type.element_type
    recv_dims = recv_type.shape

    # compute total number of elements in arrays
    send_nitems = _np.prod(send_dims, dtype=int)
    send_dtype_handle = to_dtype_handle(send_nptype)

    recv_nitems = _np.prod(recv_dims, dtype=int)
    recv_dtype_handle = to_dtype_handle(recv_nptype)

    out_types = [
        ir.RankedTensorType.get(recv_dims, recv_dtype),
        *token_type(),
    ]

    if status is None:
        status_ptr = _np.uintp(MPI_STATUS_IGNORE_ADDR)
    else:
        status_ptr = to_mpi_ptr(status)

    operands = (
        as_mhlo_constant(send_nitems, _np.intc),
        sendbuf,
        as_mhlo_constant(dest, _np.intc),
        as_mhlo_constant(sendtag, _np.intc),
        as_mhlo_constant(send_dtype_handle, _np.uintp),
        as_mhlo_constant(recv_nitems, _np.intc),
        as_mhlo_constant(source, _np.intc),
        as_mhlo_constant(recvtag, _np.intc),
        as_mhlo_constant(recv_dtype_handle, _np.uintp),
        as_mhlo_constant(to_mpi_handle(comm), _np.uintp),
        as_mhlo_constant(status_ptr, _np.uintp),
        token,
    )

    return hlo_custom_call(
        b"mpi_sendrecv",
        out_types=out_types,
        operands=operands,
        operand_layouts=get_default_layouts(operands),
        result_layouts=get_default_layouts(out_types),
        has_side_effect=True,
    )


@translation_rule_gpu
def mpi_sendrecv_xla_encode_gpu(
    ctx,
    sendbuf,
    recvbuf,
    token,
    source,
    dest,
    sendtag,
    recvtag,
    comm,
    status,
    _must_transpose=False,
):

    if _must_transpose:
        raise RuntimeError(
            "sendrecv cannot be used with forward-mode (vjp) autodiff, because "
            "the gradient might be located on a different mpi rank than the "
            "desired one. Use reverse-mode (jvp) differentiation instead."
        )

    from ..xla_bridge.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
    from ..xla_bridge.mpi_xla_bridge_gpu import build_sendrecv_descriptor

    comm = unpack_hashable(comm)
    status = unpack_hashable(status)

    send_aval, recv_aval, *_ = ctx.avals_in
    send_nptype = send_aval.dtype
    recv_nptype = recv_aval.dtype

    send_type = ir.RankedTensorType(sendbuf.type)
    send_dims = send_type.shape

    recv_type = ir.RankedTensorType(recvbuf.type)
    recv_dtype = recv_type.element_type
    recv_dims = recv_type.shape

    # compute total number of elements in arrays
    send_nitems = _np.prod(send_dims, dtype=int)
    send_dtype_handle = to_dtype_handle(send_nptype)

    recv_nitems = _np.prod(recv_dims, dtype=int)
    recv_dtype_handle = to_dtype_handle(recv_nptype)

    out_types = [
        ir.RankedTensorType.get(recv_dims, recv_dtype),
        *token_type(),
    ]

    if status is None:
        status_ptr = _np.uintp(MPI_STATUS_IGNORE_ADDR)
    else:
        status_ptr = to_mpi_ptr(status)

    operands = (
        sendbuf,
        token,
    )

    descriptor = build_sendrecv_descriptor(
        send_nitems,
        dest,
        sendtag,
        send_dtype_handle,
        recv_nitems,
        source,
        recvtag,
        recv_dtype_handle,
        to_mpi_handle(comm),
        status_ptr,
    )

    return hlo_custom_call(
        b"mpi_sendrecv",
        out_types=out_types,
        operands=operands,
        operand_layouts=get_default_layouts(operands),
        result_layouts=get_default_layouts(out_types),
        has_side_effect=True,
        backend_config=descriptor,
    )


# This function evaluates only the shapes during AST construction
def mpi_sendrecv_abstract_eval(
    sendbuf,
    recvbuf,
    token,
    source,
    dest,
    sendtag,
    recvtag,
    comm,
    status,
    _must_transpose=False,
):
    return (
        abstract_arrays.ShapedArray(recvbuf.shape, recvbuf.dtype),
        core.abstract_token,
    ), {effect}


def mpi_sendrecv_batch_eval(
    in_args,
    batch_axes,
    source,
    dest,
    sendtag,
    recvtag,
    comm,
    status,
    _must_transpose=False,
):

    sendbuf, recvbuf, token = in_args

    assert batch_axes[0] == batch_axes[1]

    res = mpi_sendrecv_p.bind(
        sendbuf,
        recvbuf,
        token,
        source=source,
        dest=dest,
        sendtag=sendtag,
        recvtag=recvtag,
        comm=comm,
        status=status,
        _must_transpose=_must_transpose,
    )
    return res, (batch_axes[0], batch_axes[2])


def mpi_sendrecv_value_and_jvp(
    in_args,
    tan_args,
    source,
    dest,
    sendtag,
    recvtag,
    comm,
    status,
    _must_transpose=False,
):
    sendbuf, recvbuf, token = in_args
    send_tan, recv_tan, token_tan = tan_args

    val, token = mpi_sendrecv_p.bind(
        sendbuf,
        recvbuf,
        token,
        source=source,
        dest=dest,
        sendtag=sendtag,
        recvtag=recvtag,
        comm=comm,
        status=status,
        _must_transpose=_must_transpose,
    )

    # throw away return token to work around jax#6285
    jvp, token_jvp = mpi_sendrecv_p.bind(
        send_tan,
        recv_tan,
        token,
        source=source,
        dest=dest,
        sendtag=sendtag,
        recvtag=recvtag,
        comm=comm,
        status=status,
        _must_transpose=not _must_transpose,
    )

    return (val, token), (jvp, ad.Zero.from_value(token_jvp))


def mpi_sendrecv_transpose_rule(
    tan_args, *x_args, source, dest, sendtag, recvtag, comm, status, _must_transpose
):
    _, _, token = x_args
    out_tan, token_tan = tan_args

    # swap the sender and receiver
    res, token = mpi_sendrecv_p.bind(
        out_tan,
        out_tan,
        token,
        source=dest,
        dest=source,
        sendtag=sendtag,
        recvtag=recvtag,
        comm=comm,
        status=status,
        _must_transpose=not _must_transpose,
    )
    return res, ad.Zero.from_value(res), token_tan


mpi_sendrecv_p.multiple_results = True
mpi_sendrecv_p.def_impl(mpi_sendrecv_impl)
mpi_sendrecv_p.def_effectful_abstract_eval(mpi_sendrecv_abstract_eval)

batching.primitive_batchers[mpi_sendrecv_p] = mpi_sendrecv_batch_eval

ad.primitive_jvps[mpi_sendrecv_p] = mpi_sendrecv_value_and_jvp
ad.primitive_transposes[mpi_sendrecv_p] = mpi_sendrecv_transpose_rule

# assign to the primitive the correct encoder
mlir.register_lowering(mpi_sendrecv_p, mpi_sendrecv_xla_encode_cpu, platform="cpu")
mlir.register_lowering(mpi_sendrecv_p, mpi_sendrecv_xla_encode_gpu, platform="cuda")



import os
import re
import warnings

import jax
import jaxlib


def versiontuple(verstr):
    # drop everything after the numeric part of the version
    allowed_chars = "0123456789."
    for i, char in enumerate(verstr):
        if char not in allowed_chars:
            break
    else:
        i = len(verstr) + 1

    verstr = verstr[:i].rstrip(".")
    return tuple(int(v) for v in verstr.split("."))[:3]


def check_jax_version():
    here = os.path.dirname(__file__)
    with open(os.path.join(here, "_latest_jax_version.txt")) as f:
        for line in f:
            line = line.strip()
            if line.startswith("#"):
                continue
            pkg, jax_maxversion = re.match(r"(\w+)\b==(.*)", line).groups()
            assert pkg == "jax"
            break

    warn_envvar = "MPI4JAX_NO_WARN_JAX_VERSION"
    nowarn = os.environ.get(warn_envvar, "").lower() in ("1", "true", "on")

    if not nowarn and versiontuple(jax.__version__) > versiontuple(jax_maxversion):
        warnings.warn(
            f"\nThe latest supported JAX version with this release of mpi4jax is {jax_maxversion}, "
            f"but you have {jax.__version__}. If you encounter problems consider downgrading JAX, "
            f"for example via:\n\n"
            f"    $ pip install jax[cpu]=={jax_maxversion}\n\n"
            f"Or try upgrading mpi4jax via\n\n"
            f"    $ pip install -U mpi4jax\n\n"
            f"You can set the environment variable `{warn_envvar}=1` to silence this warning."
        )


# TODO: remove this code once we only support jax/lib > 0.4.2
if versiontuple(jaxlib.__version__) >= (0, 4, 2):
    from jaxlib.hlo_helpers import custom_call as hlo_custom_call  # noqa: F401
else:
    from jaxlib.mhlo_helpers import custom_call as hlo_custom_call  # noqa: F401


# TODO: remove this code once we only support jax > 0.4.4
if versiontuple(jax.__version__) >= (0, 4, 4):
    from jax._src.interpreters.mlir import token_type  # noqa: F401
else:
    from jax.interpreters.mlir import token_type  # noqa: F401


# TODO: remove this code once we only support jax > 0.4.5
if versiontuple(jax.__version__) >= (0, 4, 5):

    def register_effect(EffectType):
        from jax.interpreters import mlir
        from jax._src.lax import control_flow as lcf
        import jax._src.custom_derivatives as custom_derivatives

        effect = EffectType()
        mlir.lowerable_effects.add_type(EffectType)
        lcf.allowed_effects.add_type(EffectType)
        # Effects must be added to the allow_effects list in order to work within
        # custom_vjp. See google/jax#11916
        custom_derivatives.allowed_effects.add_type(EffectType)
        return effect

else:

    def register_effect(EffectType):
        from jax.interpreters import mlir
        from jax._src.lax import control_flow as lcf
        import jax._src.custom_derivatives as custom_derivatives

        effect = EffectType()
        mlir.lowerable_effects.add(effect)
        lcf.allowed_effects.add(effect)
        # Effects must be added to the allow_effects list in order to work within
        # custom_vjp. See google/jax#11916
        custom_derivatives.allowed_effects.add(effect)
        return effect



import pytest

from mpi4py import MPI

import jax
import jax.numpy as jnp

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()


@pytest.mark.skipif(size < 2, reason="need at least 2 processes to test send/recv")
def test_send_recv():
    from mpi4jax import recv, send

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    if rank == 0:
        for proc in range(1, size):
            res, token = recv(arr, source=proc, tag=proc)
            assert jnp.array_equal(res, jnp.ones_like(arr) * proc)
            assert jnp.array_equal(_arr, arr)
    else:
        send(arr, 0, tag=rank)
        assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2, reason="need at least 2 processes to test send/recv")
def test_send_recv_scalar():
    from mpi4jax import recv, send

    arr = 1 * rank
    _arr = 1 * rank

    if rank == 0:
        for proc in range(1, size):
            res, token = recv(arr, source=proc, tag=proc)
            assert jnp.array_equal(res, jnp.ones_like(arr) * proc)
            assert jnp.array_equal(_arr, arr)
    else:
        send(arr, 0, tag=rank)
        assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2, reason="need at least 2 processes to test send/recv")
def test_send_recv_scalar_jit():
    from mpi4jax import recv, send

    arr = 1 * rank
    _arr = 1 * rank

    @jax.jit
    def send_jit(x):
        send(x, 0, tag=rank)
        return x

    if rank == 0:
        for proc in range(1, size):
            res = jax.jit(lambda x: recv(x, source=proc, tag=proc)[0])(arr)
            assert jnp.array_equal(res, jnp.ones_like(arr) * proc)
            assert jnp.array_equal(_arr, arr)
    else:
        send_jit(arr)
        assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2, reason="need at least 2 processes to test send/recv")
def test_send_recv_jit():
    from mpi4jax import recv, send

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    @jax.jit
    def send_jit(x):
        send(x, 0, tag=rank)
        return x

    if rank == 0:
        for proc in range(1, size):
            res = jax.jit(lambda x: recv(x, source=proc, tag=proc)[0])(arr)
            assert jnp.array_equal(res, jnp.ones_like(arr) * proc)
            assert jnp.array_equal(_arr, arr)
    else:
        send_jit(arr)
        assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_send_recv_deadlock():
    from mpi4jax import recv, send

    # this deadlocks without proper token management
    @jax.jit
    def deadlock(arr):
        if rank == 0:
            # send, then receive
            token = send(arr, 1)
            newarr, _ = recv(arr, 1, token=token)
        else:
            # receive, then send
            newarr, token = recv(arr, 0)
            send(arr, 0, token=token)
        return newarr

    arr = jnp.ones(10) * rank
    arr = deadlock(arr)
    assert jnp.array_equal(arr, jnp.ones_like(arr) * (1 - rank))


@pytest.mark.skipif(size < 2, reason="need at least 2 processes to test send/recv")
def test_send_recv_status():
    from mpi4jax import recv, send

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    if rank == 0:
        for proc in range(1, size):
            status = MPI.Status()
            res, token = recv(arr, source=proc, tag=proc, status=status)
            assert jnp.array_equal(res, jnp.ones_like(arr) * proc)
            assert jnp.array_equal(_arr, arr)
            assert status.Get_source() == proc
    else:
        send(arr, 0, tag=rank)
        assert jnp.array_equal(_arr, arr)


@pytest.mark.skipif(size < 2, reason="need at least 2 processes to test send/recv")
def test_send_recv_status_jit():
    from mpi4jax import recv, send

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    @jax.jit
    def send_jit(x):
        send(x, 0, tag=rank)
        return x

    if rank == 0:
        for proc in range(1, size):
            status = MPI.Status()
            res = jax.jit(lambda x: recv(x, source=proc, tag=proc, status=status)[0])(
                arr
            )
            assert jnp.array_equal(res, jnp.ones_like(arr) * proc)
            assert jnp.array_equal(_arr, arr)
            assert status.Get_source() == proc
    else:
        send_jit(arr)
        assert jnp.array_equal(_arr, arr)



from mpi4py import MPI

import jax
import jax.numpy as jnp

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()


def test_reduce():
    from mpi4jax import reduce

    arr = jnp.ones((3, 2)) * rank

    res, _ = reduce(arr, op=MPI.SUM, root=0)
    if rank == 0:
        assert jnp.array_equal(res, jnp.ones((3, 2)) * sum(range(size)))
    else:
        assert jnp.array_equal(res, arr)


def test_reduce_jit():
    from mpi4jax import reduce

    arr = jnp.ones((3, 2)) * rank

    res = jax.jit(lambda x: reduce(x, op=MPI.SUM, root=0)[0])(arr)
    if rank == 0:
        assert jnp.array_equal(res, jnp.ones((3, 2)) * sum(range(size)))
    else:
        assert jnp.array_equal(res, arr)


def test_reduce_scalar():
    from mpi4jax import reduce

    arr = rank
    res, _ = reduce(arr, op=MPI.SUM, root=0)
    if rank == 0:
        assert jnp.array_equal(res, sum(range(size)))
    else:
        assert jnp.array_equal(res, arr)


def test_reduce_scalar_jit():
    from mpi4jax import reduce

    arr = rank
    res = jax.jit(lambda x: reduce(x, op=MPI.SUM, root=0)[0])(arr)
    if rank == 0:
        assert jnp.array_equal(res, sum(range(size)))
    else:
        assert jnp.array_equal(res, arr)



from mpi4py.libmpi cimport MPI_Comm, MPI_Datatype, MPI_Status, MPI_Op

cdef int abort(int ierr, MPI_Comm comm, unicode message) nogil

cdef int abort_on_error(int ierr, MPI_Comm comm, unicode mpi_op) nogil

cdef void mpi_allgather(void* sendbuf, int sendcount, MPI_Datatype sendtype,
                        void* recvbuf, int recvcount, MPI_Datatype recvtype,
                        MPI_Comm comm) nogil

cdef void mpi_allreduce(void* sendbuf, void* recvbuf, int nitems,
                        MPI_Datatype dtype, MPI_Op op, MPI_Comm comm) nogil

cdef void mpi_alltoall(void* sendbuf, int sendcount, MPI_Datatype sendtype,
                       void* recvbuf, int recvcount, MPI_Datatype recvtype,
                       MPI_Comm comm) nogil

cdef void mpi_barrier(MPI_Comm comm) nogil

cdef void mpi_bcast(void* sendrecvbuf, int nitems, MPI_Datatype dtype,
                   int root, MPI_Comm comm) nogil

cdef void mpi_gather(void* sendbuf, int sendcount, MPI_Datatype sendtype,
                     void* recvbuf, int recvcount, MPI_Datatype recvtype,
                     int root, MPI_Comm comm) nogil

cdef void mpi_recv(void* recvbuf, int nitems, MPI_Datatype dtype, int source,
                   int tag, MPI_Comm comm, MPI_Status* status) nogil

cdef void mpi_reduce(void* sendbuf, void* recvbuf, int nitems,
                     MPI_Datatype dtype, MPI_Op op, int root,
                     MPI_Comm comm) nogil

cdef void mpi_scan(void* sendbuf, void* recvbuf, int nitems,
                   MPI_Datatype dtype, MPI_Op op, MPI_Comm comm) nogil

cdef void mpi_scatter(void* sendbuf, int sendcount, MPI_Datatype sendtype,
                      void* recvbuf, int recvcount, MPI_Datatype recvtype,
                      int root, MPI_Comm comm) nogil

cdef void mpi_send(void* sendbuf, int nitems, MPI_Datatype dtype,
                   int destination, int tag, MPI_Comm comm) nogil

cdef void mpi_sendrecv(void* sendbuf, int sendcount, MPI_Datatype sendtype, int dest, int sendtag,
                       void* recvbuf, int recvcount, MPI_Datatype recvtype, int source, int recvtag,
                       MPI_Comm comm, MPI_Status* status) nogil



import numpy as _np
from mpi4py import MPI as _MPI

from jax import abstract_arrays, core
from jax.core import Primitive, Tracer, Token
from jax.lax import create_token

from jax.interpreters import mlir
import jaxlib.mlir.ir as ir

from ..utils import (
    HashableMPIType,
    default_primitive_impl,
    to_dtype_handle,
    to_mpi_handle,
    unpack_hashable,
    wrap_as_hashable,
    as_mhlo_constant,
    get_default_layouts,
    effect,
)
from ..jax_compat import hlo_custom_call, token_type
from ..decorators import translation_rule_cpu, translation_rule_gpu
from ..validation import enforce_types
from ..comm import get_default_comm

# The Jax primitive
mpi_allgather_p = Primitive("allgather_mpi")  # Create the primitive
mpi_allgather_impl = default_primitive_impl(mpi_allgather_p)


# This function applies the primitive to an AST
@enforce_types(
    comm=(type(None), _MPI.Intracomm, HashableMPIType),
    token=(type(None), Token, Tracer),
)
def allgather(
    x,
    *,
    comm=None,
    token=None,
):
    """Perform an allgather operation.

    .. warning::

       ``x`` must have the same shape and dtype on all processes.

    Arguments:
        x: Array or scalar input to send.
        comm (mpi4py.MPI.Comm): The MPI communicator to use (defaults to
            a clone of :obj:`COMM_WORLD`).
        token (Token): XLA token to use to ensure correct execution order.
            If not given, a new token is generated.

    Returns:
        Tuple[DeviceArray, Token]:
            - Received data.
            - A new, modified token, that depends on this operation.

    """
    if token is None:
        token = create_token(x)

    if comm is None:
        comm = get_default_comm()

    comm = wrap_as_hashable(comm)

    return tuple(
        mpi_allgather_p.bind(
            x,
            token,
            comm=comm,
        )
    )


# This function compiles the operation
@translation_rule_cpu
def mpi_allgather_xla_encode_cpu(ctx, sendbuf, token, comm):
    comm = unpack_hashable(comm)

    sendbuf_aval, *_ = ctx.avals_in
    send_nptype = sendbuf_aval.dtype

    send_type = ir.RankedTensorType(sendbuf.type)
    send_dtype = send_type.element_type
    send_dims = send_type.shape

    # compute total number of elements in array
    send_nitems = _np.prod(send_dims, dtype=int)

    size = comm.Get_size()
    out_shape = (size, *send_dims)

    out_types = [
        ir.RankedTensorType.get(out_shape, send_dtype),
        *token_type(),
    ]

    operands = (
        as_mhlo_constant(send_nitems, _np.intc),
        sendbuf,
        as_mhlo_constant(to_dtype_handle(send_nptype), _np.uintp),
        # we only support matching input and output arrays
        as_mhlo_constant(send_nitems, _np.intc),
        as_mhlo_constant(to_dtype_handle(send_nptype), _np.uintp),
        #
        as_mhlo_constant(to_mpi_handle(comm), _np.uintp),
        token,
    )

    return hlo_custom_call(
        b"mpi_allgather",
        out_types=out_types,
        operands=operands,
        # layout matters here, because the first axis is special
        operand_layouts=get_default_layouts(operands, order="c"),
        result_layouts=get_default_layouts(out_types, order="c"),
        has_side_effect=True,
    )


@translation_rule_gpu
def mpi_allgather_xla_encode_gpu(ctx, sendbuf, token, comm):
    from ..xla_bridge.mpi_xla_bridge_gpu import build_allgather_descriptor

    comm = unpack_hashable(comm)

    sendbuf_aval, *_ = ctx.avals_in
    send_nptype = sendbuf_aval.dtype

    send_type = ir.RankedTensorType(sendbuf.type)
    send_dtype = send_type.element_type
    send_dims = send_type.shape

    # compute total number of elements in send array
    send_nitems = _np.prod(send_dims, dtype=int)
    send_dtype_handle = to_dtype_handle(send_nptype)

    size = comm.Get_size()
    out_shape = (size, *send_dims)

    out_types = [
        ir.RankedTensorType.get(out_shape, send_dtype),
        *token_type(),
    ]

    descriptor = build_allgather_descriptor(
        send_nitems,
        send_dtype_handle,
        # we only support matching input and output arrays
        send_nitems,
        send_dtype_handle,
        #
        to_mpi_handle(comm),
    )

    operands = (sendbuf, token)

    return hlo_custom_call(
        b"mpi_allgather",
        out_types=out_types,
        operands=operands,
        # layout matters here, because the first axis is special
        operand_layouts=get_default_layouts(operands, order="c"),
        result_layouts=get_default_layouts(out_types, order="c"),
        backend_config=descriptor,
        has_side_effect=True,
    )


# This function evaluates only the shapes during AST construction
def mpi_allgather_abstract_eval(x, token, comm):
    comm = unpack_hashable(comm)
    size = comm.Get_size()
    out_shape = (size, *x.shape)
    return (
        abstract_arrays.ShapedArray(out_shape, x.dtype),
        core.abstract_token,
    ), {effect}


mpi_allgather_p.multiple_results = True
mpi_allgather_p.def_impl(mpi_allgather_impl)
mpi_allgather_p.def_effectful_abstract_eval(mpi_allgather_abstract_eval)

mlir.register_lowering(mpi_allgather_p, mpi_allgather_xla_encode_cpu, platform="cpu")
mlir.register_lowering(mpi_allgather_p, mpi_allgather_xla_encode_gpu, platform="cuda")



"""mpi4jax demo application -- Shallow water

A non-linear shallow water solver, adapted from:

https://github.com/dionhaefner/shallow-water

Usage examples:

    # runs demo on 4 processes
    $ mpirun -n 4 python shallow_water.py

    # saves output animation as shallow-water.mp4
    $ mpirun -n 4 python shallow_water.py --save-animation

    # runs demo as a benchmark (no output)
    $ mpirun -n 4 python shallow_water.py --benchmark

"""

import os
import sys
import math
import time
import warnings
from contextlib import ExitStack
from collections import namedtuple
from functools import partial

import numpy as np
from mpi4py import MPI

try:
    import tqdm
except ImportError:
    warnings.warn("Could not import tqdm, can't show progress bar")
    HAS_TQDM = False
else:
    HAS_TQDM = True

mpi_comm = MPI.COMM_WORLD
mpi_rank = mpi_comm.Get_rank()
mpi_size = mpi_comm.Get_size()

# on GPU, put each process on its own device
os.environ["CUDA_VISIBLE_DEVICES"] = str(mpi_rank)

import jax  # noqa: E402
import jax.numpy as jnp  # noqa: E402

import mpi4jax  # noqa: E402


#
# MPI setup
#

supported_nproc = (1, 2, 4, 6, 8, 16)
if mpi_size not in supported_nproc:
    raise RuntimeError(
        f"Got invalid number of MPI processes: {mpi_size}. "
        f"Please choose one of these: {supported_nproc}."
    )

nproc_y = min(mpi_size, 2)
nproc_x = mpi_size // nproc_y

proc_idx = np.unravel_index(mpi_rank, (nproc_y, nproc_x))

#
# Grid setup
#

# we use 1 cell overlap on each side of the domain
nx_global = 360 + 2
ny_global = 180 + 2

# grid spacing in metres
dx = 5e3
dy = 5e3

# make sure processes divide the domain evenly
assert (nx_global - 2) % nproc_x == 0
assert (ny_global - 2) % nproc_y == 0

nx_local = (nx_global - 2) // nproc_x + 2
ny_local = (ny_global - 2) // nproc_y + 2

x_global, y_global = (
    np.arange(-1, nx_global - 1) * dx,
    np.arange(-1, ny_global - 1) * dy,
)
yy_global, xx_global = np.meshgrid(y_global, x_global, indexing="ij")

length_x = x_global[-2] - x_global[1]
length_y = y_global[-2] - y_global[1]

# this extracts the processor-local domain from a global array
local_slice = (
    slice((ny_local - 2) * proc_idx[0], (ny_local - 2) * proc_idx[0] + ny_local),
    slice((nx_local - 2) * proc_idx[1], (nx_local - 2) * proc_idx[1] + nx_local),
)

x_local = x_global[local_slice[1]]
y_local = y_global[local_slice[0]]

xx_local = xx_global[local_slice]
yy_local = yy_global[local_slice]

#
# Model parameters
#

# physical parameters
GRAVITY = 9.81
DEPTH = 100.0
CORIOLIS_F = 2e-4
CORIOLIS_BETA = 2e-11
CORIOLIS_PARAM = CORIOLIS_F + yy_local * CORIOLIS_BETA
LATERAL_VISCOSITY = 1e-3 * CORIOLIS_F * dx**2

# other parameters
DAY_IN_SECONDS = 86_400
PERIODIC_BOUNDARY_X = True

ADAMS_BASHFORTH_A = 1.5 + 0.1
ADAMS_BASHFORTH_B = -(0.5 + 0.1)

# output parameters
PLOT_ETA_RANGE = 10
PLOT_EVERY = 100
MAX_QUIVERS = (25, 50)


# set time step based on CFL condition
dt = 0.125 * min(dx, dy) / np.sqrt(GRAVITY * DEPTH)


@jax.jit
def get_initial_conditions():
    """For the initial conditions, we use a horizontal jet in geostrophic balance."""
    # global initial conditions
    u0_global = 10 * jnp.exp(
        -((yy_global - 0.5 * length_y) ** 2) / (0.02 * length_x) ** 2
    )
    v0_global = jnp.zeros_like(u0_global)

    # approximate balance h_y = -(f/g)u
    coriolis_global = CORIOLIS_F + yy_global * CORIOLIS_BETA
    h_geostrophy = np.cumsum(-dy * u0_global * coriolis_global / GRAVITY, axis=0)
    h0_global = (
        DEPTH
        + h_geostrophy
        # make sure h0 is centered around depth
        - h_geostrophy.mean()
        # small perturbation to break symmetry
        + 0.2
        * np.sin(xx_global / length_x * 10 * np.pi)
        * np.cos(yy_global / length_y * 8 * np.pi)
    )

    h0_local = h0_global[local_slice]
    u0_local = u0_global[local_slice]
    v0_local = v0_global[local_slice]

    token = jax.lax.create_token()
    h0_local, token = enforce_boundaries(h0_local, "h", token)
    u0_local, token = enforce_boundaries(u0_local, "u", token)
    v0_local, token = enforce_boundaries(v0_local, "v", token)

    return h0_local, u0_local, v0_local


@partial(jax.jit, static_argnums=(1,))
def enforce_boundaries(arr, grid, token=None):
    """Handle boundary exchange between processors.

    This is where mpi4jax comes in!
    """
    assert grid in ("h", "u", "v")

    # start sending west, go clockwise
    send_order = (
        "west",
        "north",
        "east",
        "south",
    )

    # start receiving east, go clockwise
    recv_order = (
        "east",
        "south",
        "west",
        "north",
    )

    overlap_slices_send = dict(
        south=(1, slice(None), Ellipsis),
        west=(slice(None), 1, Ellipsis),
        north=(-2, slice(None), Ellipsis),
        east=(slice(None), -2, Ellipsis),
    )

    overlap_slices_recv = dict(
        south=(0, slice(None), Ellipsis),
        west=(slice(None), 0, Ellipsis),
        north=(-1, slice(None), Ellipsis),
        east=(slice(None), -1, Ellipsis),
    )

    proc_neighbors = {
        "south": (proc_idx[0] - 1, proc_idx[1]) if proc_idx[0] > 0 else None,
        "west": (proc_idx[0], proc_idx[1] - 1) if proc_idx[1] > 0 else None,
        "north": (proc_idx[0] + 1, proc_idx[1]) if proc_idx[0] < nproc_y - 1 else None,
        "east": (proc_idx[0], proc_idx[1] + 1) if proc_idx[1] < nproc_x - 1 else None,
    }

    if PERIODIC_BOUNDARY_X:
        if proc_idx[1] == 0:
            proc_neighbors["west"] = (proc_idx[0], nproc_x - 1)

        if proc_idx[1] == nproc_x - 1:
            proc_neighbors["east"] = (proc_idx[0], 0)

    if token is None:
        token = jax.lax.create_token()

    for send_dir, recv_dir in zip(send_order, recv_order):
        send_proc = proc_neighbors[send_dir]
        recv_proc = proc_neighbors[recv_dir]

        if send_proc is None and recv_proc is None:
            continue

        if send_proc is not None:
            send_proc = np.ravel_multi_index(send_proc, (nproc_y, nproc_x))

        if recv_proc is not None:
            recv_proc = np.ravel_multi_index(recv_proc, (nproc_y, nproc_x))

        recv_idx = overlap_slices_recv[recv_dir]
        recv_arr = jnp.empty_like(arr[recv_idx])

        send_idx = overlap_slices_send[send_dir]
        send_arr = arr[send_idx]

        if send_proc is None:
            recv_arr, token = mpi4jax.recv(
                recv_arr, source=recv_proc, comm=mpi_comm, token=token
            )
            arr = arr.at[recv_idx].set(recv_arr)
        elif recv_proc is None:
            token = mpi4jax.send(send_arr, dest=send_proc, comm=mpi_comm, token=token)
        else:
            recv_arr, token = mpi4jax.sendrecv(
                send_arr,
                recv_arr,
                source=recv_proc,
                dest=send_proc,
                comm=mpi_comm,
                token=token,
            )
            arr = arr.at[recv_idx].set(recv_arr)

    if not PERIODIC_BOUNDARY_X and grid == "u" and proc_idx[1] == nproc_x - 1:
        arr = arr.at[:, -2].set(0.0)

    if grid == "v" and proc_idx[0] == nproc_y - 1:
        arr = arr.at[-2, :].set(0.0)

    return arr, token


ModelState = namedtuple("ModelState", "h, u, v, dh, du, dv")


@partial(jax.jit, static_argnums=(1,))
def shallow_water_step(state, is_first_step):
    """Perform one step of the shallow-water model.

    Returns modified model state.
    """
    token = jax.lax.create_token()

    h, u, v, dh, du, dv = state

    hc = jnp.pad(h[1:-1, 1:-1], 1, "edge")
    hc, token = enforce_boundaries(hc, "h", token)

    fe = jnp.empty_like(u)
    fn = jnp.empty_like(u)

    fe = fe.at[1:-1, 1:-1].set(0.5 * (hc[1:-1, 1:-1] + hc[1:-1, 2:]) * u[1:-1, 1:-1])
    fn = fn.at[1:-1, 1:-1].set(0.5 * (hc[1:-1, 1:-1] + hc[2:, 1:-1]) * v[1:-1, 1:-1])
    fe, token = enforce_boundaries(fe, "u", token)
    fn, token = enforce_boundaries(fn, "v", token)

    dh_new = dh.at[1:-1, 1:-1].set(
        -(fe[1:-1, 1:-1] - fe[1:-1, :-2]) / dx - (fn[1:-1, 1:-1] - fn[:-2, 1:-1]) / dy
    )

    # nonlinear momentum equation
    q = jnp.empty_like(u)
    ke = jnp.empty_like(u)

    # planetary and relative vorticity
    q = q.at[1:-1, 1:-1].set(
        CORIOLIS_PARAM[1:-1, 1:-1]
        + ((v[1:-1, 2:] - v[1:-1, 1:-1]) / dx - (u[2:, 1:-1] - u[1:-1, 1:-1]) / dy)
    )
    # potential vorticity
    q = q.at[1:-1, 1:-1].mul(
        1.0 / (0.25 * (hc[1:-1, 1:-1] + hc[1:-1, 2:] + hc[2:, 1:-1] + hc[2:, 2:]))
    )
    q, token = enforce_boundaries(q, "h", token)

    du_new = du.at[1:-1, 1:-1].set(
        -GRAVITY * (h[1:-1, 2:] - h[1:-1, 1:-1]) / dx
        + 0.5
        * (
            q[1:-1, 1:-1] * 0.5 * (fn[1:-1, 1:-1] + fn[1:-1, 2:])
            + q[:-2, 1:-1] * 0.5 * (fn[:-2, 1:-1] + fn[:-2, 2:])
        )
    )
    dv_new = dv.at[1:-1, 1:-1].set(
        -GRAVITY * (h[2:, 1:-1] - h[1:-1, 1:-1]) / dy
        - 0.5
        * (
            q[1:-1, 1:-1] * 0.5 * (fe[1:-1, 1:-1] + fe[2:, 1:-1])
            + q[1:-1, :-2] * 0.5 * (fe[1:-1, :-2] + fe[2:, :-2])
        )
    )
    ke = ke.at[1:-1, 1:-1].set(
        0.5
        * (
            0.5 * (u[1:-1, 1:-1] ** 2 + u[1:-1, :-2] ** 2)
            + 0.5 * (v[1:-1, 1:-1] ** 2 + v[:-2, 1:-1] ** 2)
        )
    )
    ke, token = enforce_boundaries(ke, "h", token)

    du_new = du_new.at[1:-1, 1:-1].add(-(ke[1:-1, 2:] - ke[1:-1, 1:-1]) / dx)
    dv_new = dv_new.at[1:-1, 1:-1].add(-(ke[2:, 1:-1] - ke[1:-1, 1:-1]) / dy)

    if is_first_step:
        u = u.at[1:-1, 1:-1].add(dt * du_new[1:-1, 1:-1])
        v = v.at[1:-1, 1:-1].add(dt * dv_new[1:-1, 1:-1])
        h = h.at[1:-1, 1:-1].add(dt * dh_new[1:-1, 1:-1])
    else:
        u = u.at[1:-1, 1:-1].add(
            dt
            * (
                ADAMS_BASHFORTH_A * du_new[1:-1, 1:-1]
                + ADAMS_BASHFORTH_B * du[1:-1, 1:-1]
            )
        )
        v = v.at[1:-1, 1:-1].add(
            dt
            * (
                ADAMS_BASHFORTH_A * dv_new[1:-1, 1:-1]
                + ADAMS_BASHFORTH_B * dv[1:-1, 1:-1]
            )
        )
        h = h.at[1:-1, 1:-1].add(
            dt
            * (
                ADAMS_BASHFORTH_A * dh_new[1:-1, 1:-1]
                + ADAMS_BASHFORTH_B * dh[1:-1, 1:-1]
            )
        )

    h, token = enforce_boundaries(h, "h", token)
    u, token = enforce_boundaries(u, "u", token)
    v, token = enforce_boundaries(v, "v", token)

    if LATERAL_VISCOSITY > 0:
        # lateral friction
        fe = fe.at[1:-1, 1:-1].set(
            LATERAL_VISCOSITY * (u[1:-1, 2:] - u[1:-1, 1:-1]) / dx
        )
        fn = fn.at[1:-1, 1:-1].set(
            LATERAL_VISCOSITY * (u[2:, 1:-1] - u[1:-1, 1:-1]) / dy
        )
        fe, token = enforce_boundaries(fe, "u", token)
        fn, token = enforce_boundaries(fn, "v", token)

        u = u.at[1:-1, 1:-1].add(
            dt
            * (
                (fe[1:-1, 1:-1] - fe[1:-1, :-2]) / dx
                + (fn[1:-1, 1:-1] - fn[:-2, 1:-1]) / dy
            )
        )

        fe = fe.at[1:-1, 1:-1].set(
            LATERAL_VISCOSITY * (v[1:-1, 2:] - u[1:-1, 1:-1]) / dx
        )
        fn = fn.at[1:-1, 1:-1].set(
            LATERAL_VISCOSITY * (v[2:, 1:-1] - u[1:-1, 1:-1]) / dy
        )
        fe, token = enforce_boundaries(fe, "u", token)
        fn, token = enforce_boundaries(fn, "v", token)

        v = v.at[1:-1, 1:-1].add(
            dt
            * (
                (fe[1:-1, 1:-1] - fe[1:-1, :-2]) / dx
                + (fn[1:-1, 1:-1] - fn[:-2, 1:-1]) / dy
            )
        )

    return ModelState(h, u, v, dh_new, du_new, dv_new)


@partial(jax.jit, static_argnums=(1,))
def do_multistep(state, num_steps):
    """Perform multiple model steps back-to-back."""
    return jax.lax.fori_loop(
        0, num_steps, lambda _, s: shallow_water_step(s, False), state
    )


def solve_shallow_water(t1, num_multisteps=10):
    """Iterate the model forward in time."""
    # initial conditions
    h, u, v = get_initial_conditions()
    du, dv, dh = [jnp.zeros((ny_local, nx_local)) for _ in range(3)]

    state = ModelState(h, u, v, dh, du, dv)
    sol = [state]

    state = shallow_water_step(state, True)
    sol.append(state)
    t = dt

    if HAS_TQDM:
        pbar = tqdm.tqdm(
            total=math.ceil(t1 / dt),
            disable=mpi_rank != 0,
            unit="model day",
            initial=t / DAY_IN_SECONDS,
            unit_scale=dt / DAY_IN_SECONDS,
            bar_format=(
                "{l_bar}{bar}| {n:.2f}/{total:.2f} [{elapsed}<{remaining}, "
                "{rate_fmt}{postfix}]"
            ),
        )

    # pre-compile JAX kernel
    do_multistep(state, num_multisteps)

    start = time.perf_counter()
    with ExitStack() as es:
        if HAS_TQDM:
            es.enter_context(pbar)

        while t < t1:
            state = do_multistep(state, num_multisteps)
            state[0].block_until_ready()
            sol.append(state)

            t += dt * num_multisteps

            if t < t1 and HAS_TQDM:
                pbar.update(num_multisteps)

    end = time.perf_counter()

    if mpi_rank == 0:
        print(f"\nSolution took {end - start:.2f}s")

    return sol


@jax.vmap
@jax.jit
def reassemble_array(arr):
    """This converts an array containing the solution from each processor as
    first axis to the full solution.

    Shape (mpi_size, ny_local, nx_local) -> (ny_global, nx_global)
    """
    out = jnp.empty((ny_global, nx_global), dtype=arr.dtype)
    for i in range(mpi_size):
        proc_idx_i = np.unravel_index(i, (nproc_y, nproc_x))
        local_slice_i = (
            slice(
                (ny_local - 2) * proc_idx_i[0],
                (ny_local - 2) * proc_idx_i[0] + ny_local,
            ),
            slice(
                (nx_local - 2) * proc_idx_i[1],
                (nx_local - 2) * proc_idx_i[1] + nx_local,
            ),
        )
        out = out.at[local_slice_i].set(arr[i])

    return out


def animate_shallow_water(sol):
    """Create a matplotlib animation of the result."""
    import matplotlib.pyplot as plt
    from matplotlib import animation

    quiver_stride = (
        slice(1, -1, ny_global // MAX_QUIVERS[0]),
        slice(1, -1, nx_global // MAX_QUIVERS[1]),
    )

    # set up figure
    fig = plt.figure(figsize=(6, 4))
    ax = plt.gca()

    # surface plot of height anomaly
    cs = ax.pcolormesh(
        0.5 * (x_global[:-1] + x_global[1:]) / 1e3,
        0.5 * (y_global[:-1] + y_global[1:]) / 1e3,
        sol[0].h[1:-1, 1:-1] - DEPTH,
        vmin=-PLOT_ETA_RANGE,
        vmax=PLOT_ETA_RANGE,
        cmap="RdBu_r",
    )

    # quiver plot of velocity
    cq = ax.quiver(
        xx_global[quiver_stride] / 1e3,
        yy_global[quiver_stride] / 1e3,
        sol[0].u[quiver_stride],
        sol[0].v[quiver_stride],
        clip_on=True,
    )

    # time indicator
    t = ax.text(
        s="",
        x=0.05,
        y=0.95,
        ha="left",
        va="top",
        backgroundcolor=(1, 1, 1, 0.8),
        transform=ax.transAxes,
    )

    ax.set(
        aspect="equal",
        xlim=(x_global[1] / 1e3, x_global[-2] / 1e3),
        ylim=(y_global[1] / 1e3, y_global[-2] / 1e3),
        xlabel="$x$ (km)",
        ylabel="$y$ (km)",
    )

    plt.colorbar(
        cs,
        orientation="horizontal",
        label="Surface height anomaly (m)",
        pad=0.2,
        shrink=0.8,
    )
    fig.tight_layout()

    def animate(i):
        state = sol[i]

        eta = state.h - DEPTH
        cs.set_array(eta[1:-1, 1:-1].flatten())
        cq.set_UVC(state.u[quiver_stride], state.v[quiver_stride])

        current_time = PLOT_EVERY * dt * i
        t.set_text(f"t = {current_time / DAY_IN_SECONDS:.2f} days")
        return (cs, cq, t)

    anim = animation.FuncAnimation(
        fig, animate, frames=len(sol), interval=50, blit=True, repeat_delay=3_000
    )
    return anim


if __name__ == "__main__":
    benchmark_mode = "--benchmark" in sys.argv

    sol = solve_shallow_water(t1=10 * DAY_IN_SECONDS, num_multisteps=PLOT_EVERY)

    if benchmark_mode:
        sys.exit(0)

    # copy solution to mpi_rank 0
    full_sol_arr, _ = mpi4jax.gather(jnp.asarray(sol), root=0, comm=mpi_comm)

    if mpi_rank == 0:
        # full_sol_arr has shape (nproc, time, nvars, ny, nx)
        full_sol_arr = jnp.moveaxis(full_sol_arr, 0, 2)
        full_sol = [ModelState(*reassemble_array(x)) for x in full_sol_arr]

        anim = animate_shallow_water(full_sol)

        if "--save-animation" in sys.argv:
            # save animation as MP4 video (requires ffmpeg)
            anim.save("shallow-water.mp4", writer="ffmpeg", dpi=100)
        else:
            import matplotlib.pyplot as plt

            plt.show()



from mpi4py import MPI

import jax
import jax.numpy as jnp

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()


def test_scan():
    from mpi4jax import scan

    arr = jnp.ones((3, 2)) * rank

    res, _ = scan(arr, op=MPI.SUM)
    assert jnp.array_equal(res, jnp.ones((3, 2)) * sum(range(rank + 1)))


def test_scan_jit():
    from mpi4jax import scan

    arr = jnp.ones((3, 2)) * rank

    res = jax.jit(lambda x: scan(x, op=MPI.SUM)[0])(arr)
    assert jnp.array_equal(res, jnp.ones((3, 2)) * sum(range(rank + 1)))


def test_scan_scalar():
    from mpi4jax import scan

    arr = rank
    res, _ = scan(arr, op=MPI.SUM)
    assert jnp.array_equal(res, sum(range(rank + 1)))


def test_scan_scalar_jit():
    from mpi4jax import scan

    arr = rank
    res = jax.jit(lambda x: scan(x, op=MPI.SUM)[0])(arr)
    assert jnp.array_equal(res, sum(range(rank + 1)))



cdef extern from "cuda_runtime_api.h":
    cdef enum cudaError:
        cudaSuccess = 0

    cdef enum cudaMemcpyKind:
        cudaMemcpyHostToHost = 0
        cudaMemcpyHostToDevice = 1
        cudaMemcpyDeviceToHost = 2
        cudaMemcpyDeviceToDevice = 3
        cudaMemcpyDefault = 4

    ctypedef cudaError cudaError_t

    ctypedef void* cudaStream_t
    cudaError_t cudaStreamSynchronize(cudaStream_t stream) nogil
    cudaError_t cudaMemcpy(void *dst, const void *src, size_t count, cudaMemcpyKind kind) nogil

    const char* cudaGetErrorName(cudaError_t error) nogil
    const char* cudaGetErrorString(cudaError_t error) nogil



import numpy as _np
from mpi4py import MPI as _MPI

from jax import core
from jax.core import Primitive, Tracer, Token
from jax.lax import create_token

from jax.interpreters import mlir
import jaxlib.mlir.ir as ir

from ..utils import (
    HashableMPIType,
    default_primitive_impl,
    to_dtype_handle,
    to_mpi_handle,
    unpack_hashable,
    wrap_as_hashable,
    as_mhlo_constant,
    get_default_layouts,
    effect,
)
from ..jax_compat import hlo_custom_call, token_type
from ..decorators import translation_rule_cpu, translation_rule_gpu
from ..validation import enforce_types
from ..comm import get_default_comm


# The Jax primitive
mpi_send_p = Primitive("send_mpi")  # Create the primitive
mpi_send_impl = default_primitive_impl(mpi_send_p)


# This function applies the primitive to an AST
@enforce_types(
    dest=_np.integer,
    tag=_np.integer,
    comm=(type(None), _MPI.Intracomm, HashableMPIType),
    token=(type(None), Token, Tracer),
)
def send(x, dest, *, tag=0, comm=None, token=None):
    """Perform a send operation.

    Arguments:
        x: Array or scalar input to send.
        dest (int): Rank of the destination MPI process.
        tag (int): Tag of this message.
        comm (mpi4py.MPI.Comm): The MPI communicator to use (defaults to
            a clone of :obj:`COMM_WORLD`).
        token (Token): XLA token to use to ensure correct execution order.
            If not given, a new token is generated.

    Returns:
        Token: A new, modified token, that depends on this operation.

    """
    if token is None:
        token = create_token(x)

    if comm is None:
        comm = get_default_comm()

    comm = wrap_as_hashable(comm)
    return mpi_send_p.bind(x, token, dest=dest, tag=tag, comm=comm)


# This function compiles the operation
@translation_rule_cpu
def mpi_send_xla_encode_cpu(ctx, x, token, dest, tag, comm):
    comm = unpack_hashable(comm)

    x_aval, *_ = ctx.avals_in
    x_nptype = x_aval.dtype

    x_type = ir.RankedTensorType(x.type)
    dims = x_type.shape

    # compute total number of elements in array
    nitems = _np.prod(dims, dtype=int)
    dtype_handle = to_dtype_handle(x_nptype)

    out_types = token_type()

    operands = (
        as_mhlo_constant(nitems, _np.intc),
        x,
        as_mhlo_constant(dest, _np.intc),
        as_mhlo_constant(tag, _np.intc),
        as_mhlo_constant(to_mpi_handle(comm), _np.uintp),
        as_mhlo_constant(dtype_handle, _np.uintp),
        token,
    )

    # JAX insists on outputs being iterable
    return [
        hlo_custom_call(
            b"mpi_send",
            out_types=out_types,
            operands=operands,
            operand_layouts=get_default_layouts(operands),
            result_layouts=get_default_layouts(out_types),
            has_side_effect=True,
        )
    ]


@translation_rule_gpu
def mpi_send_xla_encode_gpu(ctx, x, token, dest, tag, comm):
    from ..xla_bridge.mpi_xla_bridge_gpu import build_send_descriptor

    comm = unpack_hashable(comm)

    x_aval, *_ = ctx.avals_in
    x_nptype = x_aval.dtype

    x_type = ir.RankedTensorType(x.type)
    dims = x_type.shape

    # compute total number of elements in array
    nitems = _np.prod(dims, dtype=int)
    dtype_handle = to_dtype_handle(x_nptype)

    out_types = token_type()

    operands = (
        x,
        token,
    )

    descriptor = build_send_descriptor(
        nitems,
        dest,
        tag,
        to_mpi_handle(comm),
        dtype_handle,
    )

    # JAX insists on outputs being iterable
    return [
        hlo_custom_call(
            b"mpi_send",
            out_types=out_types,
            operands=operands,
            operand_layouts=get_default_layouts(operands),
            result_layouts=get_default_layouts(out_types),
            has_side_effect=True,
            backend_config=descriptor,
        )
    ]


# This function evaluates only the shapes during AST construction
def mpi_send_abstract_eval(xs, token, dest, tag, comm):
    return core.abstract_token, {effect}


mpi_send_p.def_impl(mpi_send_impl)
mpi_send_p.def_effectful_abstract_eval(mpi_send_abstract_eval)

# assign to the primitive the correct encoder
mlir.register_lowering(mpi_send_p, mpi_send_xla_encode_cpu, platform="cpu")
mlir.register_lowering(mpi_send_p, mpi_send_xla_encode_gpu, platform="cuda")



import pytest


@pytest.fixture(autouse=True)
def add_examples_to_path():
    import os
    import sys

    rootpath = os.path.dirname(__file__)
    demopath = os.path.join(rootpath, "..", "examples")
    sys.path.append(demopath)

    try:
        yield
    finally:
        removed_path = sys.path.pop()
        assert removed_path == demopath


def test_shallow_water_demo():
    from shallow_water import solve_shallow_water

    sol = solve_shallow_water(86_400)
    assert len(sol) > 100



import tempfile
import time
import os

from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()


def cap_to_file(capsys, write_to):
    out = capsys.readouterr().out
    with open(write_to, "a") as f:
        f.write(out)


def test_barrier(capsys):
    """Verify that barrier blocks execution by printing messages before and after"""
    from mpi4jax._src.flush import flush
    from mpi4jax import barrier

    # pipe all messages to the same file
    tmpdir = tempfile.gettempdir()

    write_to = os.path.join(tmpdir, "mpi4jax-barrier.txt")
    if rank == 0:
        with open(write_to, "w"):
            pass

    print(f"r{rank} | start")

    time.sleep(rank * 0.2)
    cap_to_file(capsys, write_to)

    # without a barrier here, some ranks would start writing
    # "done" before everyone has writen "start"
    token = barrier()  # noqa: F841
    flush()

    print(f"r{rank} | done")

    time.sleep(rank * 0.2)
    cap_to_file(capsys, write_to)

    time.sleep(size * 0.2)

    with open(write_to, "r") as f:
        outputs = f.readlines()

    assert len(outputs) == size * 2
    assert all(o.endswith("start\n") for o in outputs[:size])
    assert all(o.endswith("done\n") for o in outputs[size:])



import pytest

from mpi4py import MPI

import jax
import jax.numpy as jnp
import numpy as np

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

from mpi4jax._src.jax_compat import versiontuple  # noqa: E402

pytestmark = pytest.mark.xfail(
    versiontuple(jax.__version__) >= versiontuple("0.4.4"),
    reason="does not work with jax>=0.4.4 yet",
    raises=ImportError,
)


def test_allreduce():
    from mpi4jax import allreduce
    from mpi4jax.experimental import auto_tokenize

    arr = jnp.ones((3, 2))

    def f(a):
        res, _ = allreduce(a, op=MPI.SUM)
        res, _ = allreduce(res, op=MPI.SUM)
        return res

    res = auto_tokenize(f)(arr)
    np.testing.assert_allclose(res, arr * size**2)


def test_nested_jits():
    from mpi4jax import allreduce
    from mpi4jax.experimental import auto_tokenize

    arr = jnp.ones((3, 2))

    @jax.jit
    def f(a):
        res, _ = allreduce(a, op=MPI.SUM)

        @jax.jit
        def g(b):
            res, _ = allreduce(b, op=MPI.SUM)
            res, _ = allreduce(res, op=MPI.SUM)
            return res

        res = g(res)
        res = g(res)
        return res

    res = auto_tokenize(f)(arr)
    np.testing.assert_allclose(res, arr * size**5)


@pytest.mark.skipif(size < 2, reason="need 2 processes")
def test_send_recv_tokenizer():
    from mpi4jax import recv, send
    from mpi4jax.experimental import auto_tokenize

    def simple_message_pass(arr):
        if rank == 0:
            b, _ = recv(arr, source=1, comm=comm)
            return b
        elif rank == 1:
            a = arr + 1
            send(a, dest=0, comm=comm)
            return arr
        return jnp.zeros_like(arr)

    res = jax.jit(auto_tokenize(simple_message_pass))(jnp.zeros((2, 2)))
    if rank == 0:
        np.testing.assert_allclose(res, jnp.ones((2, 2)))
    if rank == 1:
        np.testing.assert_allclose(res, jnp.zeros((2, 2)))


@pytest.mark.skipif(size < 2, reason="need 2 processes")
def test_send_recv_hotpotato_tokenizer():
    from mpi4jax import recv, send, barrier
    from mpi4jax.experimental import auto_tokenize

    def hot_potato(arr):
        # Here, we test a send/recv pattern that is extremely likely to return the
        # wrong result unless the order is preserved.
        barrier()  # Free barrier test aswell.
        if rank == 0:
            a = arr + 1
            b, _ = recv(arr, source=1, comm=comm)
            send(a, dest=1, comm=comm)
            c, _ = recv(arr, source=1, comm=comm)
            d, _ = recv(arr, source=1, comm=comm)
            e = c + d
            send(e, dest=1, comm=comm)
            f, _ = recv(arr, source=1, comm=comm)
            send(e + f, dest=1, comm=comm)
            send(e * d, dest=1, comm=comm)
            return f
        elif rank == 1:
            a = arr + 2
            send(a, dest=0, comm=comm)
            b, _ = recv(arr, source=0, comm=comm)
            send(a + b, dest=0, comm=comm)
            send(a * b, dest=0, comm=comm)
            c, _ = recv(arr, source=0, comm=comm)
            send(b - c, dest=0, comm=comm)
            d, _ = recv(arr, source=0, comm=comm)
            e, _ = recv(arr, source=0, comm=comm)
            return d + e
        return jnp.zeros_like(arr)

    # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # ! Proof that this test actually works !
    # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    #
    # Uncomment the below line of code.
    # The final result will be wrong and the asserts will fail.
    #
    # ---------------------------
    # auto_tokenize = lambda x: x
    # ---------------------------

    real_result = hot_potato(jnp.zeros((2, 2)))
    jitted_tokenized = jax.jit(auto_tokenize(jax.jit(hot_potato)))(jnp.zeros((2, 2)))
    np.testing.assert_allclose(real_result, jitted_tokenized)
    # Assert to ourselves the results are correct.
    if rank == 0:
        np.testing.assert_allclose(jitted_tokenized, jnp.ones((2, 2)) * -4)
    if rank == 1:
        np.testing.assert_allclose(jitted_tokenized, jnp.ones((2, 2)) * 11)


def test_fori_loop_tokenizer():
    from mpi4jax import allreduce
    from mpi4jax.experimental import auto_tokenize

    NUM_LOOPS = 6

    def sum_loop(i, args):
        (arr,) = args
        res, _ = allreduce(arr, op=MPI.SUM)
        return [res]

    def my_method(arr):
        return jax.lax.fori_loop(0, NUM_LOOPS, sum_loop, [arr])

    res = jax.jit(auto_tokenize(my_method))(jnp.ones((2, 2)))
    np.testing.assert_allclose(res[0], np.ones((2, 2)) * size**NUM_LOOPS)


def test_while_loop_tokenizer():
    from mpi4jax import allreduce
    from mpi4jax.experimental import auto_tokenize

    def sum_loop(arr):
        res, _ = allreduce(arr, op=MPI.SUM)
        res = res + 1
        return res

    def cond(arr):
        return jnp.all(arr < 1000)

    def my_method(arr):
        return jax.lax.while_loop(cond, sum_loop, arr)

    res = jax.jit(auto_tokenize(my_method))(jnp.ones((2, 2)))
    assert (res >= np.ones((2, 2)) * 1000).all()


def test_cond_tokenizer():
    from mpi4jax import allreduce
    from mpi4jax.experimental import auto_tokenize

    def branch1(arr):
        res, _ = allreduce(arr, op=MPI.PROD)
        return res

    def branch2(arr):
        res, _ = allreduce(arr, op=MPI.SUM)
        return res

    def my_method(bool, arr):
        return jax.lax.cond(bool, branch1, branch2, arr)

    res1 = jax.jit(auto_tokenize(my_method))(
        jnp.asarray(True), jnp.ones((2, 2), dtype=jnp.int32)
    ).block_until_ready()
    res2 = jax.jit(auto_tokenize(my_method))(
        jnp.asarray(False), jnp.ones((2, 2), dtype=jnp.int32)
    ).block_until_ready()
    assert (res1 == 1).all()
    assert (res2 == size).all()


def test_allgather_scalar():
    from mpi4jax import allgather
    from mpi4jax.experimental import auto_tokenize

    @jax.jit
    @auto_tokenize
    def f(arr):
        res, _ = allgather(arr)
        return res

    res = f(jnp.asarray(rank))
    assert jnp.array_equal(res, jnp.arange(size))


def test_alltoall_jit():
    from mpi4jax import alltoall
    from mpi4jax.experimental import auto_tokenize

    arr = jnp.ones((size, 3, 2)) * rank

    res = jax.jit(auto_tokenize(lambda x: alltoall(x)[0]))(arr)
    for p in range(size):
        assert jnp.array_equal(res[p], jnp.ones((3, 2)) * p)


def test_bcast_scalar_jit():
    from mpi4jax import bcast
    from mpi4jax.experimental import auto_tokenize

    arr = 1
    _arr = 1

    if rank != 0:
        _arr = _arr * 0

    res = jax.jit(auto_tokenize(lambda x: bcast(x, root=0)[0]))(_arr)
    assert jnp.array_equal(res, arr)
    if rank == 0:
        assert jnp.array_equal(_arr, arr)


def test_gather_scalar_jit():
    from mpi4jax import gather
    from mpi4jax.experimental import auto_tokenize

    arr = rank
    res = jax.jit(auto_tokenize(lambda x: gather(x, root=0)[0]))(arr)
    if rank == 0:
        assert jnp.array_equal(res, jnp.arange(size))
    else:
        assert jnp.array_equal(res, arr)


def test_reduce_scalar_jit():
    from mpi4jax import reduce
    from mpi4jax.experimental import auto_tokenize

    arr = rank
    res = jax.jit(auto_tokenize(lambda x: reduce(x, op=MPI.SUM, root=0)[0]))(arr)
    if rank == 0:
        assert jnp.array_equal(res, sum(range(size)))
    else:
        assert jnp.array_equal(res, arr)


def test_scan_scalar_jit():
    from mpi4jax import scan
    from mpi4jax.experimental import auto_tokenize

    arr = rank
    res = jax.jit(auto_tokenize(lambda x: scan(x, op=MPI.SUM)[0]))(arr)
    assert jnp.array_equal(res, sum(range(rank + 1)))


def test_scatter_jit():
    from mpi4jax import scatter
    from mpi4jax.experimental import auto_tokenize

    if rank == 0:
        arr = jnp.stack([jnp.ones((3, 2)) * r for r in range(size)], axis=0)
    else:
        arr = jnp.ones((3, 2)) * rank

    res = jax.jit(auto_tokenize(lambda x: scatter(x, root=0)[0]))(arr)
    assert jnp.array_equal(res, jnp.ones((3, 2)) * rank)


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_sendrecv_status_jit():
    from mpi4jax import sendrecv
    from mpi4jax.experimental import auto_tokenize

    arr = jnp.ones((3, 2)) * rank
    _arr = arr.copy()

    other = 1 - rank

    status = MPI.Status()
    res = jax.jit(
        auto_tokenize(
            lambda x, y: sendrecv(x, y, source=other, dest=other, status=status)[0]
        )
    )(arr, arr)

    assert jnp.array_equal(res, jnp.ones_like(arr) * other)
    assert jnp.array_equal(_arr, arr)
    assert status.Get_source() == other


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_while_loop_consistency():
    from mpi4jax import allreduce, sendrecv, barrier
    from mpi4jax.experimental import auto_tokenize

    def cond(value):
        barrier()
        if rank == 1 or rank == 0:
            new_value, _ = sendrecv(
                value, jnp.zeros_like(value), source=1 - rank, dest=1 - rank
            )
        return jnp.all(new_value < 10)

    def loop(value):
        barrier()
        new_value, _ = allreduce(value, MPI.SUM)
        return new_value

    @jax.jit
    @auto_tokenize
    def run(value):
        return jax.lax.while_loop(cond, loop, value)

    res = run(1.0)
    assert res >= 10


@pytest.mark.skipif(size < 2 or rank > 1, reason="Runs only on rank 0 and 1")
def test_cond_consistency():
    from mpi4jax import recv, send
    from mpi4jax.experimental import auto_tokenize

    def hot_potato_right(args):
        arr, t = args
        a = arr + 1
        b, t = recv(arr, source=1, comm=comm, token=t)
        t = send(a, dest=1, comm=comm, token=t)
        c, t = recv(arr, source=1, comm=comm, token=t)
        d, t = recv(arr, source=1, comm=comm, token=t)
        e = c + d
        t = send(e, dest=1, comm=comm, token=t)
        f, t = recv(arr, source=1, comm=comm, token=t)
        t = send(e + f, dest=1, comm=comm, token=t)
        t = send(e * d, dest=1, comm=comm, token=t)
        return f, t

    def hot_potato_left(args):
        arr, t = args
        a = arr + 2
        t = send(a, dest=0, comm=comm, token=t)
        b, t = recv(arr, source=0, comm=comm, token=t)
        t = send(a + b, dest=0, comm=comm, token=t)
        t = send(a * b, dest=0, comm=comm, token=t)
        c, t = recv(arr, source=0, comm=comm, token=t)
        t = send(b - c, dest=0, comm=comm, token=t)
        d, t = recv(arr, source=0, comm=comm, token=t)
        e, t = recv(arr, source=0, comm=comm, token=t)
        return d + e, t

    @jax.jit
    @auto_tokenize
    def run(arr, my_rank):
        t = jax.lax.create_token()
        res1, t = jax.lax.cond(
            my_rank == 0, hot_potato_right, hot_potato_left, (arr, t)
        )
        return res1

    res = run(jnp.zeros((2, 2)), rank)

    if rank == 0:
        expected = -4
    else:
        expected = 11

    np.testing.assert_allclose(res, jnp.ones((2, 2)) * expected)
